'''
Created on 17 feb. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.io.GpkgWriter import GpkgWriter


class AbstractGeoDataFrameDemos(object):
    '''
    classdocs
    '''

    @staticmethod
    def postprocess(sio, crs='epsg:2154'):
        raise Exception('Deprecated!')

    @staticmethod
    def _dump(mapOfGdf, gpkgOutputFile="/tmp/dump.gpkg"):
        if (not mapOfGdf is None) and (0 < len(mapOfGdf)):
            GpkgWriter(mapOfGdf, gpkgOutputFile).run()



'''
Created on 12 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from pandas import concat
from shapely.geometry import Point
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraphFactory import UrbanGraphFactory


class STShortestPath(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, roads, fromPoints, toPoints):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, "GeoDataFrame")
        if not isinstance(fromPoints, (GeoDataFrame, Point)):
            raise IllegalArgumentTypeException(
                fromPoints, "GeoDataFrame or single Point")
        if not isinstance(toPoints, (GeoDataFrame, Point)):
            raise IllegalArgumentTypeException(
                toPoints, "GeoDataFrame or single Point")

        self.graph = UrbanGraphFactory.create(roads, method="networkx")

        self.roads = roads

        if isinstance(fromPoints, GeoDataFrame):
            self.fromPoints = list(fromPoints.centroid)
        else:
            self.fromPoints = [fromPoints]

        if isinstance(toPoints, GeoDataFrame):
            self.toPoints = list(toPoints.centroid)
        else:
            self.toPoints = [toPoints]

    def run(self):
        gdfs = []
        for fromPoint in self.fromPoints:
            for toPoint in self.toPoints:
                gdfs.append(self.graph.shortest_path(fromPoint, toPoint))

        result = GeoDataFrame(concat(gdfs), crs=self.roads.crs)
        result["gid"] = range(len(result))
        return result



'''
Created on 29 nov. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from geopandas import GeoDataFrame
from numpy import full
from t4gpd.commons.TypeLib import TypeLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess
from t4gpd.raster.RTFromArrayToRaster import RTFromArrayToRaster


class STFromGridToRaster(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, grid, fieldname, rowFieldname="row", columnFieldname="column",
                 roi=None, ndv=-1, debug=False):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt

        if not isinstance(grid, GeoDataFrame):
            raise IllegalArgumentTypeException(grid, "GeoDataFrame")
        self.grid = grid

        for field in [fieldname, rowFieldname, columnFieldname]:
            if not field in grid:
                raise IllegalArgumentTypeException(
                    field, "must be a valid grid field name")
        self.fieldname = fieldname
        self.rowFieldname = rowFieldname
        self.columnFieldname = columnFieldname
        self.roi = roi
        self.ndv = ndv
        self.debug = debug

    def __get_nrows_ncols(self):
        # if self.roi is None:
        nrows = 1 + self.grid[self.rowFieldname].max()
        ncols = 1 + self.grid[self.columnFieldname].max()
        return nrows, ncols

    def __as_array(self, nrows, ncols):
        arr = full([nrows, ncols], self.ndv, dtype=type(self.ndv))

        if not TypeLib.are_both_floating_or_integer(self.ndv, self.grid[self.fieldname]):
            warnings.warn(
                "The numerical types of the NDV and the series of values are not identical")

        for _, row in self.grid.iterrows():
            nr = row[self.rowFieldname]
            nc = row[self.columnFieldname]
            value = row[self.fieldname]
            arr[(nrows - 1) - nr, nc] = value

        return arr

    def run(self):
        nrows, ncols = self.__get_nrows_ncols()
        arr = self.__as_array(nrows, ncols)
        # print(arr.shape)
        # return arr

        roi = self.grid if self.roi is None else self.roi
        raster = RTFromArrayToRaster(arr, roi, self.ndv, self.debug).run()
        return raster


"""
import matplotlib.pyplot as plt
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.raster.RTToFile import RTToFile

buildings = GeoDataFrameDemos.ensaNantesBuildings()
buildings = GeoDataFrameDemos.singleBuildingInNantes()
# buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
# buildings = GeoDataFrameDemos.regularGridOfPlots(2, 3)
grid = STGrid(buildings, dx=5, dy=5, indoor="both", intoPoint=True,
              encode=True, withDist=False).run()

ndv = 0
# grid.loc[len(grid), ["row", "column", "indoor"]] = int(grid.row.max() + 6), int(grid.column.max()), ndv
grid.row = grid.row.astype(int)
grid.column = grid.column.astype(int)

raster = STFromGridToRaster(grid, "indoor", ndv=ndv, debug=False).run()

fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
buildings.plot(ax=ax, color="lightgrey")
grid.plot(ax=ax, color="red")
ax.axis("off")
plt.tight_layout()
plt.show()

RTToFile(raster, "./raster.tif").run()
plt.imshow(raster.read(1), cmap="BrBG")
plt.show()
plt.close(fig)
"""



'''
Created on 30 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from geopandas import GeoDataFrame
from numpy import asarray, cos, dot, linspace, mean, pi, sin, sum
from pandas import concat
from shapely import LineString, LinearRing, Point, Polygon, get_parts, union_all
from shapely.affinity import scale, translate
from shapely.ops import substring
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.overlap.OverlapLib import OverlapLib
from t4gpd.commons.raycasting.PrepareMasksLib import PrepareMasksLib
from t4gpd.commons.proj.DoubleProjectionLib import DoubleProjectionLib
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.commons.raycasting.SurroundingMasks25DLib import SurroundingMasks25DLib


class DoubleProjectionOfWallsLib(object):
    '''
    classdocs
    '''
    CIRCLE = LinearRing(zip(cos(linspace(0, 2*pi, 90)),
                        sin(linspace(0, 2*pi, 90))))
    NO_IDX = -123456789
    warnings.formatwarning = WarnUtils.format_Warning_alt

    @staticmethod
    def __project_2d_point_onto_unit_circle(viewpoint, p, circle):
        curv_absc = circle.project(Point(p), normalized=True)
        pp = circle.interpolate(curv_absc, normalized=True)
        return [pp.x, pp.y, viewpoint[2]], curv_absc

    @staticmethod
    def __double_projection_of_LineString(prj, viewpoint, lineStringMask,
                                          size, npts, circle):
        circle = translate(circle, xoff=viewpoint.x, yoff=viewpoint.y)

        pts = [lineStringMask.interpolate(
            x, normalized=True).coords[0] for x in linspace(0, 1, npts)]

        viewpoint = viewpoint.coords[0]
        ppts = [prj(viewpoint, p, size) for p in pts]
        z0 = mean([p.z for p in ppts])

        pA, curvA = DoubleProjectionOfWallsLib.__project_2d_point_onto_unit_circle(
            viewpoint, ppts[0], circle)
        pB, curvB = DoubleProjectionOfWallsLib.__project_2d_point_onto_unit_circle(
            viewpoint, ppts[-1], circle)

        if (curvA < curvB):
            if (0.9 <= 1 - curvB + curvA):
                # TL 13.03.2024: Debug DoubleProjectionOfWallsLibTest.testWalls3()
                warnings.warn(
                    f"Issue with viewpoint={viewpoint} and {lineStringMask.wkt}")
                return None
            arc1 = substring(circle,
                             start_dist=curvB,
                             end_dist=1,
                             normalized=True)
            arc2 = substring(circle,
                             start_dist=0,
                             end_dist=curvA,
                             normalized=True)
            arcPts = [(p[0], p[1], viewpoint[2]) for p in arc1.coords] +\
                [(p[0], p[1], viewpoint[2]) for p in arc2.coords]
        else:
            arc = substring(circle,
                            start_dist=curvB,
                            end_dist=curvA,
                            normalized=True)
            arcPts = [(p[0], p[1], viewpoint[2]) for p in arc.coords]

        return GeomLib.forceZCoordinateToZ0(Polygon(ppts + arcPts), z0=z0)

    @staticmethod
    def __foo(reverse_prj, size, masks, rayLength):
        vppks = sorted(masks.__VIEWPOINT_PK__.unique())

        masks4 = []
        for vppk in vppks:
            masks2 = masks.loc[masks[vppk ==
                                     masks.__VIEWPOINT_PK__].index,
                               ["__MASK_IDS__", "__VIEWPOINT_PK__", "__VIEWPOINT_GEOM__", "__WALL_MASK__", "geometry"]]
            vpgeom = masks2.__VIEWPOINT_GEOM__.tolist()[0]

            masks3 = OverlapLib.overlap(
                masks2, fieldnames=["__MASK_IDS__", "__WALL_MASK__"], patchid="patch_id")
            masks3["__VIEWPOINT_PK__"] = vppk
            masks3["__VIEWPOINT_GEOM__"] = vpgeom
            masks3["__REPRESENTATIVE_POINT__"] = masks3.geometry.apply(
                lambda geom: GeomLib.forceZCoordinateToZ0(geom.representative_point(), z0=vpgeom.z))
            masks3["__MASK_AREA__"] = masks3.geometry.apply(
                lambda geom: geom.area)
            masks4.append(masks3)

        masks4 = GeoDataFrame(concat(masks4), crs=masks.crs)
        masks4.__REPRESENTATIVE_POINT__ = masks4.apply(
            lambda row: reverse_prj(
                row.__VIEWPOINT_GEOM__, row.__REPRESENTATIVE_POINT__, size),
            axis=1)
        masks4["__1ST_FRONT__"] = masks4.apply(lambda row: DoubleProjectionOfWallsLib.__managing_overlapping_fronts(
            row.__VIEWPOINT_GEOM__, row.__REPRESENTATIVE_POINT__, row.__WALL_MASK__, rayLength), axis=1)

        def __get_id(ids, mid): return f"ERR_{vppk}" if (
            DoubleProjectionOfWallsLib.NO_IDX == mid) else ids[mid]
        masks4["__MASK_DIST__"] = masks4.__1ST_FRONT__.apply(lambda v: v[0])
        masks4["__MASK_ID__"] = masks4.apply(lambda row: __get_id(
            row.__MASK_IDS__, row.__1ST_FRONT__[1]), axis=1)

        def __get_wall_mask(
            wall_masks, mid): return LineString([]) if (DoubleProjectionOfWallsLib.NO_IDX == mid) else wall_masks[mid]
        masks4["__WALL_MASK__"] = masks4.apply(lambda row: __get_wall_mask(
            row.__WALL_MASK__, row.__1ST_FRONT__[1]), axis=1)

        masks4 = masks4.dissolve(by=["__VIEWPOINT_PK__", "__MASK_ID__", "__WALL_MASK__"], aggfunc={
                                 "__VIEWPOINT_GEOM__": "first", "__MASK_DIST__": list,
                                 "__MASK_AREA__": list}, as_index=True)

        def __get_dist(dists, areas): return dot(
            asarray(areas), asarray(dists)) / sum(asarray(areas))
        masks4["__DEPTH__"] = masks4.apply(lambda row: __get_dist(
            row.__MASK_DIST__, row.__MASK_AREA__), axis=1)

        masks4.reset_index(drop=False, inplace=True)
        masks4.drop(columns=["__MASK_DIST__", "__MASK_AREA__"], inplace=True)
        return masks4

    @staticmethod
    def __managing_overlapping_fronts(viewpoint, representative_point, wall_masks, rayLength):
        vp, rp, d = viewpoint, representative_point, viewpoint.distance(
            representative_point)
        ray = LineString([
            (vp.x, vp.y),
            (vp.x + (rp.x - vp.x) * (rayLength / d),
             vp.y + (rp.y - vp.y) * (rayLength / d)),
        ])

        minDist, minIdx = float("inf"), DoubleProjectionOfWallsLib.NO_IDX
        for idx, __WALL_MASK__ in enumerate(wall_masks):
            dist = vp.distance(ray.intersection(__WALL_MASK__))
            if (dist < minDist):
                minDist, minIdx = dist, idx

        # if minIdx is None:
        #     for idx, __WALL_MASK__ in enumerate(wall_masks):
        #         dist = ray.distance(__WALL_MASK__)
        #         if (dist < minDist):
        #             minDist, minIdx = dist, idx

        return minDist, minIdx
        # return minIdx

    @staticmethod
    def walls(sensors, buildings, maskidFieldname="ID", elevationFieldname="HAUTEUR",
              horizon=100.0, h0=0.0, size=1, projectionName="Stereographic", npts=5,
              aggregate=True, encode=True):
        if not GeoDataFrameLib.shareTheSameCrs(buildings, sensors):
            raise Exception(
                "Illegal argument: buildings and sensors must share shames CRS!")

        rayLength = GeoDataFrameLib.getBBoxDiagonal(buildings).length

        prj = DoubleProjectionLib.projection_switch(projectionName)
        reverse_prj = DoubleProjectionLib.reverse_projection_switch(
            projectionName)

        sensors2 = sensors.copy(deep=True)
        sensors2.geometry = sensors2.geometry.apply(
            lambda geom: geom if geom.has_z else GeomLib.forceZCoordinateToZ0(geom, h0))

        # PRE-PROCESSING
        masks = buildings.assign(__MASK_IDS__=lambda row: row[maskidFieldname])
        masks = PrepareMasksLib.removeNonVisible25DMasks(
            sensors2, masks, elevationFieldname, horizon, h0
        )
        # masks.to_csv("/tmp/1.csv", index=False, sep=";")  # DEBUG
        # masks = SurroundingMasks25DLib.removeHidden25DMasks(masks)
        # masks.to_csv("/tmp/2.csv", index=False, sep=";")  # DEBUG

        circle = scale(DoubleProjectionOfWallsLib.CIRCLE,
                       xfact=size, yfact=size, origin=(0, 0))

        # PROCESSING
        masks = masks.explode(ignore_index=True)  # Debug 23.08.2028
        masks["__WALL_MASK__"] = masks.loc[:, "geometry"]
        # masks.to_csv("/tmp/3.csv", index=False, sep=";")
        masks.geometry = masks.apply(
            lambda row: DoubleProjectionOfWallsLib.__double_projection_of_LineString(
                prj, row.__VIEWPOINT_GEOM__, row.geometry, size=size, npts=npts, circle=circle),
            axis=1)
        # masks.to_csv("/tmp/4.csv", index=False, sep=";")

        masks = DoubleProjectionOfWallsLib.__foo(
            reverse_prj, size, masks, rayLength)
        # masks.to_csv("/tmp/5.csv", index=False, sep=";")

        # POST-PROCESSING
        # masks.drop(columns=["__VIEWPOINT__"], inplace=True)
        if aggregate:
            masks.__VIEWPOINT_GEOM__ = masks.__VIEWPOINT_GEOM__.apply(
                lambda g: g.wkt)
            masks = masks.dissolve(
                by=["__VIEWPOINT_PK__", "__VIEWPOINT_GEOM__"], as_index=True)
            masks.geometry = masks.geometry.apply(
                lambda geom: geom.buffer(0))
            masks.reset_index(drop=False, inplace=True)
            masks.drop(
                columns=["__MASK_ID__", "__WALL_MASK__", "__DEPTH__"], inplace=True)

            if ("isoaire" == projectionName.lower()):
                tmp = pi * size**2
                masks["svf"] = masks.geometry.apply(
                    lambda geom: geom.area / tmp)
        else:
            masks = masks.explode(ignore_index=True)
            masks.rename(columns={"__DEPTH__": "depth"}, inplace=True)
            masks.depth = masks.depth.apply(
                lambda v: horizon if (float("inf") == v) else v)

        if encode:
            if "__REPRESENTATIVE_POINT__" in masks:
                masks.__REPRESENTATIVE_POINT__ = masks.__REPRESENTATIVE_POINT__.apply(
                    lambda g: g.wkt)
            if "__WALL_MASK__" in masks:
                masks.__WALL_MASK__ = masks.__WALL_MASK__.apply(
                    lambda g: g.wkt)

        return masks

    def test(ofile=None):
        import matplotlib.pyplot as plt
        from geopandas import clip
        from matplotlib_scalebar.scalebar import ScaleBar
        from shapely.geometry import box
        from t4gpd.commons.proj.DoubleProjectionOfPointsLib import DoubleProjectionOfPointsLib
        from t4gpd.demos.GeoDataFrameDemosC import GeoDataFrameDemosC
        from t4gpd.morph.STGrid import STGrid

        iris = GeoDataFrameDemosC.irisTastavin()
        buildings = GeoDataFrameDemosC.irisTastavinBuildings()
        # buildings.ID = buildings.ID.apply(lambda v: v[-6:-1])
        streetlights = GeoDataFrameDemosC.irisTastavinStreetlights()

        streetlights.geometry = streetlights.geometry.apply(
            lambda geom: GeomLib.forceZCoordinateToZ0(geom, 12))

        dx = 150
        grid = STGrid(buildings, dx=dx, dy=None, indoor=False, intoPoint=True,
                      encode=True, withDist=False).execute()  # < 15 sec
        sensors = clip(grid, iris, keep_geom_type=True)
        # sensors = sensors[sensors.gid == 13]

        size = 4
        # projectionName = "Isoaire"
        # projectionName="Orthogonal"
        projectionName = "Stereographic"
        pp = DoubleProjectionOfPointsLib.points(
            sensors, streetlights, h0=0,
            size=size, projectionName=projectionName, encode=True)
        pmasks = DoubleProjectionOfWallsLib.walls(
            sensors, buildings, maskidFieldname="ID", elevationFieldname="HAUTEUR",
            horizon=100.0, h0=0.0, size=size, projectionName=projectionName, npts=5,
            aggregate=False, encode=True)

        if not ofile is None:
            iris.to_file(ofile, layer="iris")
            buildings.to_file(ofile, layer="buildings")
            streetlights.to_file(ofile, layer="streetlights")
            sensors.to_file(ofile, layer="sensors")
            pp.to_file(ofile, layer="pp")
            pmasks.to_file(ofile, layer="pmasks")

        # minx, miny, maxx, maxy = box(
        #     *sensors[sensors.gid == 6].total_bounds).buffer(6.0).bounds
        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.35 * 8.26))
        ax.set_title(
            f"Montpellier (34), IRIS Tastavin ({projectionName})", size=28)
        iris.boundary.plot(ax=ax, color="red", linestyle="dotted")
        buildings.plot(ax=ax, color="grey")
        sensors.plot(ax=ax, color="black", marker="P")
        sensors.buffer(size).boundary.plot(
            ax=ax, color="black", linewidth=0.3, linestyle=":")
        pmasks.plot(ax=ax, column="depth", cmap="Spectral",
                    edgecolor="black", alpha=0.75, legend="True")
        # pp.plot(ax=ax, column="depth", cmap="Spectral",
        #         marker="+", legend="True")
        ax.axis("off")
        # ax.axis([minx, maxx, miny, maxy])
        scalebar = ScaleBar(dx=1.0, units="m", length_fraction=None, box_alpha=0.85,
                            width_fraction=0.005, location="lower right", frameon=True)
        ax.add_artist(scalebar)
        fig.tight_layout()
        plt.show()
        plt.close(fig)

        return pp, pmasks


# pp, pmasks = DoubleProjectionOfWallsLib.test("/tmp/DoubleProjectionOfWallsLib.gpkg")



'''
Created on 25 nov. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import arange, cos, pi, sin
from pandas.core.common import flatten
from shapely import union_all
from shapely.geometry import box, CAP_STYLE, LineString, MultiLineString, Point, Polygon
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.PolarCartesianCoordinates import PolarCartesianCoordinates


class Pos(object):
    UNKNOWN = -1
    START = 0
    MID = 1
    STOP = 2


class ExactIsovistLib(object):
    '''
    classdocs
    '''
    EPSILON = 1e-2

    @staticmethod
    def __c2p(viewpoint, linestring):
        a, b = linestring.coords
        ar, at = PolarCartesianCoordinates.fromCartesianToPolarCoordinates(
            a[0] - viewpoint.x, a[1] - viewpoint.y)
        br, bt = PolarCartesianCoordinates.fromCartesianToPolarCoordinates(
            b[0] - viewpoint.x, b[1] - viewpoint.y)
        if pi > abs(bt - at):
            if (at < bt):
                return linestring, (ar, at), (br, bt)
            return LineString([b, a]), (br, bt), (ar, at)
        if (at < bt):
            return LineString([b, a]), (br, bt - 2 * pi), (ar, at)
        return linestring, (ar, at - 2 * pi), (br, bt)

    @staticmethod
    def __fromCartesianToPolarRadiusEquation(viewpoint, linestring):
        a, b = linestring.coords
        x1, y1 = a[0] - viewpoint.x, a[1] - viewpoint.y
        x2, y2 = b[0] - viewpoint.x, b[1] - viewpoint.y
        den = x1 * y2 - x2 * y1
        if (0.0 == den):
            # The straight line goes through the origin
            raise Exception('Unreachable instruction!')
        # Linear equation: ax + by = 1
        a, b = (y2 - y1) / den, (x1 - x2) / den
        return (lambda theta: (1.0 / (a * cos(theta) + b * sin(theta))))

    @staticmethod
    def __isHiddenBy(viewpoint, curr, other):
        # Returns True if 'other' is hidden by 'curr' when observing from 'viewpoint'
        cr1, ct1, cr2, ct2 = curr['r1'], curr['t1'], curr['r2'], curr['t2']
        or1, ot1, or2, ot2 = other['r1'], other['t1'], other['r2'], other['t2']

        if ((ct1 <= ot1 <= ot2 <= ct2) or
                ((ct1 < 0) and (360 + ct1 <= ot1 <= ot2 <= 360 + ct2))):
            if (max(cr1, cr2) < min(or1, or2)):
                return True
            a, b = other['geometry'].coords
            if (LineString([viewpoint.coords[0][0:2], a[0:2]]).intersects(curr['geometry']) and
                    LineString([viewpoint.coords[0][0:2], b[0:2]]).intersects(curr['geometry'])):
                return True
        return False

    @staticmethod
    def __decimationOfOpaqueEdges(buildings, viewpoint, maxRayLength):
        # Selection of buildings located in the region of interest
        roi = viewpoint.buffer(maxRayLength, cap_style=CAP_STYLE.square).bounds
        # ids = buildings.sindex.intersection(roi)
        # masks = [buildings.loc[_id].geometry for _id in ids]
        masks = list(filter(box(*roi).intersection, buildings.geometry))

        # Use a buffer to avoid slivers
        # masks = [mask.buffer(1e-3, join_style=JOIN_STYLE.mitre) for mask in masks]

        # Geometric union of buildings to remove party walls
        masks = union_all(masks)
        masks = masks.geoms if GeomLib.isMultipart(masks) else [masks]

        # Remove non-visible edges (badly oriented)
        masks = [GeomLib.normalizeRingOrientation(m) for m in masks]
        masks = list(
            flatten([GeomLib.toListOfBipointsAsLineStrings(m) for m in masks]))
        masks = [m for m in masks if GeomLib.isInFrontOf(viewpoint, m)]

        # From Cartesian to Polar coordinates (in the coordinates system centered on the viewpoint)
        masks = [{'tmp': ExactIsovistLib.__c2p(viewpoint, m)} for m in masks]

        masks = [{
            'geometry': m['tmp'][0],
            'r1': m['tmp'][1][0],
            't1': AngleLib.toDegrees(m['tmp'][1][1]),
            'r2': m['tmp'][2][0],
            't2': AngleLib.toDegrees(m['tmp'][2][1]),
        } for m in masks]

        # Sort edges according to angles (t1) and radii (r1)
        masks.sort(key=lambda m: (m['t1'], m['r1']), reverse=False)
        '''
        for i, m in enumerate(masks):  # DEBUG
            print(f"{i}:: t1: {m['t1']}, t2: {m['t2']}")  # DEBUG
        '''

        # Remove completely hidden edges (phase 1)
        hiddenEdgeIds, nEdges = set(), len(masks)
        for i, curr in enumerate(masks):
            if (not i in hiddenEdgeIds):
                for j in range(i + 1, nEdges):
                    if (not j in hiddenEdgeIds):
                        other = masks[j]
                        if ExactIsovistLib.__isHiddenBy(viewpoint, curr, other):
                            hiddenEdgeIds.add(j)

        for i in sorted(hiddenEdgeIds, reverse=True):
            del (masks[i])

        '''
        if (0 < len(masks)):  # DEBUG
            GeoDataFrame(masks, crs=buildings.crs).to_file('./x/__dbg1.shp')  # DEBUG
        '''

        # Remove completely hidden edges (phase 2)
        hiddenEdgeIds, nEdges = set(), len(masks)
        for i, curr in enumerate(masks):
            if (not i in hiddenEdgeIds):
                fin, stack, stackOfIdx = -1, [curr['geometry'].coords[0]], [i]
                for j in range(i + 1, nEdges):
                    if ((not j in hiddenEdgeIds) and
                            (curr['r2'] == masks[j]['r1']) and (curr['t2'] == masks[j]['t1'])):
                        fin, maxDist = j, max(
                            curr['r1'], curr['r2'], masks[j]['r1'], masks[j]['r2'])
                        stack.append(masks[j]['geometry'].coords[1])
                        stackOfIdx.append(j)
                        for k in range(j + 1, nEdges):
                            if ((not k in hiddenEdgeIds) and
                                    (masks[fin]['r2'] == masks[k]['r1']) and (masks[fin]['t2'] == masks[k]['t1'])):
                                fin, maxDist = k, max(
                                    maxDist, masks[k]['r1'], masks[k]['r2'])
                                stack.append(masks[k]['geometry'].coords[1])
                                stackOfIdx.append(k)

                if (-1 < fin):
                    fin, stack = masks[fin], LineString(stack)
                    for j in range(i + 1, nEdges):
                        if (not j in hiddenEdgeIds) and (j not in stackOfIdx):
                            other = masks[j]
                            if ((curr['t1'] <= other['t1'] <= other['t2'] <= fin['t2']) or
                                    ((360 + curr['t1'] < 0) and (curr['t1'] <= other['t1'] <= other['t2'] <= 360 + fin['t2']))):
                                if (maxDist < min(other['r1'], other['r2'])):
                                    hiddenEdgeIds.add(j)
                                else:
                                    a, b = other['geometry'].coords
                                    if (LineString([viewpoint.coords[0][0:2], a[0:2]]).intersects(stack) and
                                            LineString([viewpoint.coords[0][0:2], b[0:2]]).intersects(stack)):
                                        hiddenEdgeIds.add(j)

        for i in sorted(hiddenEdgeIds, reverse=True):
            del (masks[i])

        masks = [(x.update({'edge_id': i}), x)[1] for i, x in enumerate(masks)]

        '''
        if (0 < len(masks)):  # DEBUG
            GeoDataFrame(masks, crs=buildings.crs).to_file('./x/__dbg2.shp')  # DEBUG
        '''
        return GeoDataFrame(masks, crs=buildings.crs)

    @staticmethod
    def __fromOpaqueEdgesToNodes(edges, viewpoint, maxRayLength):
        # 1st step: put both ends of each opaque edge into a common "bag" (opaqueNodes)
        opaqueNodes = dict()
        for _, row in edges.iterrows():
            edgeId, r1, t1, r2, t2, geometry = row[[
                'edge_id', 'r1', 't1', 'r2', 't2', 'geometry']]

            # CONSIDER WHETHER maxRayLength < max(r1, r2)
            if (r1 <= maxRayLength) and (r2 <= maxRayLength):
                _pairOfNodes = [(r1, t1, Pos.START), (r2, t2, Pos.STOP)]
            else:
                _tmp = GeomLib.splitSegmentAccordingToTheDistanceToViewpoint(
                    geometry, viewpoint, maxRayLength)

                if ((r1 <= maxRayLength) and (maxRayLength < r2) and (1 == len(_tmp))):
                    _pairOfNodes = [geometry.coords[0][0:2], _tmp[0]]
                elif ((maxRayLength < r1) and (r2 <= maxRayLength) and (1 == len(_tmp))):
                    _pairOfNodes = [_tmp[0], geometry.coords[1][0:2]]
                elif ((maxRayLength < r1) and (maxRayLength < r2) and (2 == len(_tmp))):
                    _pairOfNodes = _tmp
                else:
                    _pairOfNodes = []

                if (0 < len(_pairOfNodes)):
                    _tmp = ExactIsovistLib.__c2p(
                        viewpoint, LineString(_pairOfNodes))
                    geometry = _tmp[0]
                    r1, t1 = _tmp[1][0], AngleLib.toDegrees(_tmp[1][1])
                    r2, t2 = _tmp[2][0], AngleLib.toDegrees(_tmp[2][1])
                    _pairOfNodes = [(r1, t1, Pos.START), (r2, t2, Pos.STOP)]
                    for fieldname, fieldvalue in [('r1', r1), ('t1', t1), ('r2', r2), ('t2', t2), ('geometry', geometry)]:
                        row[fieldname] = fieldvalue

            for r, t, pos in _pairOfNodes:
                _geom = geometry.coords[0] if (
                    Pos.START == pos) else geometry.coords[-1]
                t = t if (t >= 0) else (360 + t)
                _curr = {'edge_id': edgeId, 'r': r, 't': t,
                         'geometry': Point(_geom), 'pos': pos}
                if (t in opaqueNodes):
                    if (r in opaqueNodes[t]):
                        opaqueNodes[t][r].append(_curr)
                    else:
                        opaqueNodes[t][r] = [_curr]
                else:
                    opaqueNodes[t] = {r: [_curr]}

        # 2nd step: for each opaque edge, identify the set of covered azimuths
        for _, row in edges.iterrows():
            edgeId, t1, t2, linestring = row[[
                'edge_id', 't1', 't2', 'geometry']]
            _equ = ExactIsovistLib.__fromCartesianToPolarRadiusEquation(
                viewpoint, linestring)

            for t, _mapOfNodes in opaqueNodes.items():
                edgeIds = []
                for _listOfNodes in _mapOfNodes.values():
                    edgeIds += [_node['edge_id'] for _node in _listOfNodes]
                if (((t1 < t < t2) or ((t1 < 0) and (t1 < t - 360 < t2))) and
                        (edgeId not in edgeIds)):
                    _t_rad = AngleLib.toRadians(t)
                    r = _equ(_t_rad)
                    if (r < maxRayLength + ExactIsovistLib.EPSILON):
                        def _ppt(r, t): return Point(
                            (viewpoint.x + r * cos(t), viewpoint.y + r * sin(t)))
                        _curr = {'edge_id': edgeId, 'r': r, 't': t,
                                 'geometry': _ppt(r, _t_rad), 'pos': Pos.UNKNOWN}
                        if r in opaqueNodes[t]:
                            opaqueNodes[t][r].append(_curr)
                        else:
                            opaqueNodes[t][r] = [_curr]

        '''
        if (0 < len(opaqueNodes)):
            rows = []  # DEBUG
            for _k in opaqueNodes.keys():  # DEBUG
                for _kk in opaqueNodes[_k]:  # DEBUG
                    rows += opaqueNodes[_k][_kk]  # DEBUG
            GeoDataFrame(rows, crs='epsg:2154').to_file('./x/__dbg3.shp')  # DEBUG
    
            rows2 = []  # DEBUG
            for _row in rows:  # DEBUG
                _geom = LineString([ (viewpoint.x, viewpoint.y), (_row['geometry'].x, _row['geometry'].y)])  # DEBUG
                rows2.append({  # DEBUG
                    'geometry': _geom, 'edge_id': _row['edge_id'],  # DEBUG
                    'r': _row['r'], 't': _row['t'], 'pos': _row['pos']  # DEBUG
                    })  # DEBUG
            GeoDataFrame(rows2, crs='epsg:2154').to_file('./x/__dbg4.gpkg')  # DEBUG
        '''

        return opaqueNodes

    @staticmethod
    def __decimationOfOpaqueNodes(nodes, viewpoint, maxRayLength):
        opaqueNodes = dict()

        # 1st step: remove hidden nodes
        for t in nodes.keys():
            opaqueNodes[t] = dict()
            for r in sorted(nodes[t].keys()):
                nNodes = len(nodes[t][r])

                if (1 == nNodes):
                    currNode = nodes[t][r][0]
                    currNode['edge_id'] = str(currNode['edge_id'])
                    opaqueNodes[t][r] = currNode
                    if (Pos.UNKNOWN == currNode['pos']):
                        # do not store all nodes beyond an intermediate node
                        opaqueNodes[t][r]['pos'] = Pos.MID
                        break
                    elif ((Pos.START == currNode['pos']) and Epsilon.equals(currNode['r'], maxRayLength, ExactIsovistLib.EPSILON)):
                        # do not store nodes beyond an extreme one
                        break

                elif (2 == nNodes):
                    n1, n2 = nodes[t][r]

                    # TODO: add one more test regarding possible edge alignment
                    if (((n1['pos'] in [Pos.STOP, Pos.UNKNOWN]) and (Pos.START == n2['pos'])) or
                            ((Pos.STOP == n1['pos']) and (n2['pos'] in [Pos.START, Pos.UNKNOWN]))):
                        _ids = ArrayCoding.encode(
                            [n1['edge_id'], n2['edge_id']])
                    elif (((Pos.START == n1['pos']) and (n2['pos'] in [Pos.STOP, Pos.UNKNOWN])) or
                          ((n1['pos'] in [Pos.START, Pos.UNKNOWN]) and (Pos.STOP == n2['pos']))):
                        _ids = ArrayCoding.encode(
                            [n2['edge_id'], n1['edge_id']])
                    else:
                        raise Exception('Unreachable instruction!')

                    opaqueNodes[t][r] = {
                        'edge_id': _ids, 'r': r, 't': t, 'geometry': n1['geometry'], 'pos': Pos.MID}
                    # do not store all nodes beyond a pair of start and end nodes
                    break

                else:
                    raise Exception('Unreachable instruction!')

        '''
        if (0 < len(opaqueNodes)):  # DEBUG
            rows = []  # DEBUG
            for _k in opaqueNodes.keys():  # DEBUG
                for _kk in opaqueNodes[_k]:  # DEBUG
                    rows.append(opaqueNodes[_k][_kk])  # DEBUG
            GeoDataFrame(rows, crs='epsg:2154').to_file('./x/__dbg5.shp')  # DEBUG
        '''

        # 2nd step: build the ordered list of contour points
        ctrPts = []
        for t in sorted(opaqueNodes.keys()):
            _nNodes = len(opaqueNodes[t])
            if 0 < _nNodes:
                _curr = list(opaqueNodes[t].values())[0]

                if (1 == _nNodes):
                    ctrPts.append(_curr)

                else:
                    if (Pos.STOP == _curr['pos']):
                        _rev = False
                    elif (Pos.START == _curr['pos']):
                        _rev = True
                    else:
                        raise Exception('Unreachable instruction!')
                    for r in sorted(opaqueNodes[t], reverse=_rev):
                        ctrPts.append(opaqueNodes[t][r])

        '''
        if (0 < len(ctrPts)):  # DEBUG
            GeoDataFrame(ctrPts, crs='epsg:2154').to_file('./x/__dbg6.shp')  # DEBUG
        '''

        return ctrPts

    @staticmethod
    def __addAnArtificialPoint(viewpoint, maxRayLength, theta):
        _t_rad = AngleLib.toRadians(theta)
        _geom = Point((
            viewpoint.x + maxRayLength * cos(_t_rad),
            viewpoint.y + maxRayLength * sin(_t_rad)))
        return {
            'edge_id': None,
            'r': maxRayLength,
            't': theta % 360,
            'geometry': _geom,
            'pos': Pos.UNKNOWN
        }

    @staticmethod
    def __addSetOfArtificialPoints(viewpoint, maxRayLength, theta1, theta2, delta):
        result = []
        for _t in arange(theta1, theta2, delta):
            result.append(ExactIsovistLib.__addAnArtificialPoint(
                viewpoint, maxRayLength, _t))
        if (_t < theta2):
            result.append(ExactIsovistLib.__addAnArtificialPoint(
                viewpoint, maxRayLength, theta2))
        return result

    @staticmethod
    def __addAnArtificialHorizon(nodes, viewpoint, maxRayLength, delta):
        ctrPts = []
        if (0 == len(nodes)):
            ctrPts = ExactIsovistLib.__addSetOfArtificialPoints(
                viewpoint, maxRayLength, 0, 360, delta)

        elif (2 == len(nodes)):
            if (Pos.START == nodes[0]['pos']) and (Pos.STOP == nodes[1]['pos']):
                _start, _stop = nodes
            elif (Pos.STOP == nodes[0]['pos']) and (Pos.START == nodes[1]['pos']):
                _stop, _start = nodes
            else:
                raise Exception('Unreachable instruction!')

            _t0 = _stop['t']
            _t1 = (_start['t'] + 360) if (_stop['t']
                                          > _start['t']) else _start['t']

            ctrPts.append(_stop)
            ctrPts += ExactIsovistLib.__addSetOfArtificialPoints(
                viewpoint, maxRayLength, _t0, _t1, delta)
            ctrPts.append(_start)

        else:
            for i, _curr in enumerate(nodes):
                _prev = nodes[i - 1]
                if ((_prev['edge_id'] != _curr['edge_id']) and
                    ((Pos.STOP == _prev['pos']) or ((Pos.MID == _prev['pos']) and (Epsilon.equals(_prev['r'], maxRayLength, ExactIsovistLib.EPSILON)))) and
                        ((Pos.START == _curr['pos']) or ((Pos.MID == _curr['pos']) and (Epsilon.equals(_curr['r'], maxRayLength, ExactIsovistLib.EPSILON))))):

                    _t0 = _prev['t']
                    _t1 = (_curr['t']) if (_prev['t'] <
                                           _curr['t']) else (_curr['t'] + 360)

                    ctrPts += ExactIsovistLib.__addSetOfArtificialPoints(
                        viewpoint, maxRayLength, _t0, _t1, delta)

                ctrPts.append(_curr)
        return ctrPts

    @staticmethod
    def __anticipation(viewpoint, occludingSurfaces):
        _anticip = 0
        for _occl in occludingSurfaces.geoms:
            _FgBg = _occl.length
            _VpFg, _VpBg = [viewpoint.distance(Point(c)) for c in _occl.coords]
            _VpFg, _VpBg = min(_VpFg, _VpBg), max(_VpFg, _VpBg)
            _anticip += (_VpBg * _FgBg) / _VpFg
        return _anticip

    @staticmethod
    def __buildTheIsovist(viewpoint, nodes, crs):
        def __intersects(l1, l2): return 0 < len(
            set(l1.split('#')).intersection(set(l2.split('#'))))

        _isovist = Polygon([(p['geometry'].x, p['geometry'].y) for p in nodes])
        _isovist_perim = _isovist.length

        # Artificial Horizon, material and occluding surfaces
        _artificialHorizon = []
        _materialSurfaces = []
        _occludingSurfaces = []

        for i, _curr in enumerate(nodes):
            _prev = nodes[i - 1]

            if (_prev['edge_id'] is None) and (_curr['edge_id'] is None):
                _edge = LineString((_prev['geometry'], _curr['geometry']))
                _artificialHorizon.append(_edge)

            elif ((_prev['edge_id'] is not None) and (_curr['edge_id'] is not None)):
                _a = GeomLib.removeZCoordinate(_prev['geometry'])
                _b = GeomLib.removeZCoordinate(_curr['geometry'])
                _edge = LineString((_a, _b))

                if __intersects(_prev['edge_id'], _curr['edge_id']):
                    _materialSurfaces.append(_edge)
                else:
                    _occludingSurfaces.append(_edge)

            else:
                _a = GeomLib.removeZCoordinate(_prev['geometry'])
                _b = GeomLib.removeZCoordinate(_curr['geometry'])
                _edge = LineString((_a, _b))
                _occludingSurfaces.append(_edge)

        _artificialHorizon = MultiLineString(_artificialHorizon)
        _skyline_1 = _artificialHorizon.length
        _skyline_2 = _skyline_1 / _isovist_perim

        _materialSurfaces = MultiLineString(_materialSurfaces)
        _solidness_1 = _materialSurfaces.length
        _solidness_2 = _solidness_1 / _isovist_perim

        _occludingSurfaces = MultiLineString(_occludingSurfaces)
        _occlusivity_1 = _occludingSurfaces.length
        _occlusivity_2 = _occlusivity_1 / _isovist_perim

        _isovist = {
            'geometry': _isovist,
            'perim': _isovist_perim,
            'occlusiv': _occlusivity_1,
            'occlusiv_r': _occlusivity_2,
            'solid': _solidness_1,
            'solid_r': _solidness_2,
            'skyline': _skyline_1,
            'skyline_r': _skyline_2,
            'anticipati': ExactIsovistLib.__anticipation(viewpoint, _occludingSurfaces),
            'nodes': nodes,
            'artif_hori': _artificialHorizon,
            'solid_surf': _materialSurfaces,
            'occlu_surf': _occludingSurfaces,
        }
        return _isovist

    @staticmethod
    def run(buildings, viewpoint, maxRayLength, delta=10.0):
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(
                buildings, 'buildings GeoDataFrame')
        if not isinstance(viewpoint, Point):
            raise IllegalArgumentTypeException(viewpoint, 'Point')

        edges = ExactIsovistLib.__decimationOfOpaqueEdges(
            buildings, viewpoint, maxRayLength)
        nodes = ExactIsovistLib.__fromOpaqueEdgesToNodes(
            edges, viewpoint, maxRayLength)
        nodes = ExactIsovistLib.__decimationOfOpaqueNodes(
            nodes, viewpoint, maxRayLength)
        nodes = ExactIsovistLib.__addAnArtificialHorizon(
            nodes, viewpoint, maxRayLength, delta)

        return ExactIsovistLib.__buildTheIsovist(viewpoint, nodes, buildings.crs)



'''
Created on 2 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from hashlib import md5
from os import path


class Checksum(object):
    '''
    classdocs
    '''

    @staticmethod
    def md5(filename):
        if path.exists(filename):
            return md5(open(filename, 'rb').read()).hexdigest()
        return None



'''
Created on 23 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''


class EnergyLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def fromLongWaveFluxToSurfaceTemperature(lw):
        '''
        Excerpt from https://doi.org/10.1016/j.buildenv.2024.112176

        Surface temperatures are a more efficient index than raw flux 
        values for assessing the radiative specificities of a position
        in the infrared domain. They are derived from long-wave fluxes
        using the equation ST=(LW/(5.670374419*10^(-8)))^(1/4)-273.15.
        '''
        Tsurf = (lw / 5.670374419e-8)**(1/4) - 273.15
        return Tsurf

    @staticmethod
    def plot(ofile=None):
        import matplotlib.pyplot as plt
        from numpy import linspace

        lw = linspace(100, 1100, 100)
        Tsurf = EnergyLib.fromLongWaveFluxToSurfaceTemperature(lw)

        fig, ax = plt.subplots(figsize=(1.4 * 8.26, 1 * 8.26))
        ax.plot(lw, Tsurf)
        ax.set_xlabel("Long-wave flux [W/m$^2$]", fontsize=16)
        ax.set_ylabel("Surface temperature [°C]", fontsize=16)
        ax.grid()
        fig.tight_layout()
        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)


# EnergyLib.plot()



'''
Created on 22 nov. 2020

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import line_interpolate_point, LineString, Point, union_all
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.graph.UrbanGraphFactory import UrbanGraphFactory


class NeighborhoodLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def hash_edge(id1, id2):
        str_hash = ArrayCoding.encode([int(min(id1, id2)), int(max(id1, id2))])
        return str_hash

    @staticmethod
    def neighborhood(roads, fromPoint, maxDist):
        # WRONG IDEA:
        # roi = fromPoint.buffer(maxDist, cap_style=CAP_STYLE.round)
        # roi = GeoDataFrame([{"geometry": roi}], crs=roads.crs)
        # ug = UrbanGraphFactory.create(roads, method="networkx", roi=roi)
        ug = UrbanGraphFactory.create(roads, method="networkx")
        sourceIdx = ug._add_new_point(fromPoint)

        # Identification of full edges and vertices in the neighborhood
        burned_vertices, burned_edges = {sourceIdx: maxDist}, {}
        for vtxIdx, vtxGeom in ug.icVertices.items():
            geom = Point(vtxGeom)
            if (maxDist >= fromPoint.distance(geom)):
                path = ug.shortest_path(fromPoint, geom)
                if (0 < len(path)):
                    pathVertices = ArrayCoding.decode(path.path.squeeze())
                    pathLen = path.geometry.squeeze().length

                    if (maxDist >= pathLen):
                        burned_vertices[vtxIdx] = maxDist - pathLen
                        for i in range(1, len(pathVertices)):
                            burned_edges[NeighborhoodLib.hash_edge(
                                pathVertices[i-1], pathVertices[i]
                            )] = True

        rows = []
        for str_hash in burned_edges.keys():
            id1, id2 = ArrayCoding.decode(str_hash, outputType=int)
            rows.append(LineString([ug.icVertices[id1], ug.icVertices[id2]]))

        # Identification of remaining edges in the neighborhood
        for id1, remaining_dist in burned_vertices.items():
            for id2 in ug.nx_graph.adj[id1].keys():
                str_hash = NeighborhoodLib.hash_edge(id1, id2)
                if not str_hash in burned_edges:
                    ls = LineString([ug.icVertices[id1], ug.icVertices[id2]])
                    rp = line_interpolate_point(ls, remaining_dist)
                    ls = LineString([ug.icVertices[id1], rp])
                    rows.append(ls)

        return GeoDataFrame([{"geometry": union_all(rows)}], crs=roads.crs)



'''
Created on 30 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely import MultiPolygon, Polygon, union_all
from shapely.prepared import prep
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ConcaveLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def fillInTheConcavities(polygon, obstacles):
        if not GeomLib.isPolygonal(polygon):
            raise IllegalArgumentTypeException(
                polygon, "Polygon or MultiPolygon")

        chull = polygon.convex_hull
        concavities = chull.difference(polygon)
        if isinstance(concavities, Polygon):
            concavities = MultiPolygon([concavities])

        pobstacles = prep(obstacles)
        result = [polygon] + \
            list(filter(pobstacles.disjoint, concavities.geoms))

        return union_all(result)



'''
Created on 12 sep. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.overlap.OverlapLib import OverlapLib


class STMultipleOverlaps(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, pks, patchid="__PATCH_ID__"):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        self.gdf = gdf

        self.pks = pks
        self.patchid = patchid

    def run(self):
        return OverlapLib.overlap(self.gdf, self.pks, patchid=self.patchid)



'''
Created on 20 nov. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import asarray
from shapely import LineString
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error, mean_squared_error
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class LinearRegressionIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, with_geom=False):
        self.with_geom = with_geom

    def runWithArgs(self, row):
        pts = GeomLib.getListOfShapelyPoints(
            row.geometry, withoutClosingLoops=True)

        X = asarray([pt.x for pt in pts]).reshape(-1, 1)
        y = asarray([pt.y for pt in pts])
        reg = LinearRegression().fit(X, y)

        y_pred = reg.predict(X)
        slope, intercept = reg.coef_[0], reg.intercept_
        score = reg.score(X, y)
        mae = mean_absolute_error(y, y_pred)
        mse = mean_squared_error(y, y_pred)

        if self.with_geom:
            X_minmax = asarray([X.min(), X.max()]).reshape(-1, 1)
            y_minmax_pred = reg.predict(X_minmax)
            geom = LineString(
                [(X_minmax[0, 0], y_minmax_pred[0]), (X_minmax[1, 0], y_minmax_pred[0])])

            return {
                "geometry": geom,
                "slope": slope,
                "intercept": intercept,
                "score": score,
                "mae": mae,
                "mse": mse
            }
        return {
            "slope": slope,
            "intercept": intercept,
            "score": score,
            "mae": mae,
            "mse": mse
        }


"""
import matplotlib.pyplot as plt
from geopandas import GeoDataFrame
from shapely import box
from shapely.affinity import rotate
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess

gpd = GeoDataFrame({"geometry": [rotate(box(0, 0, 10, 10), angle=30)]})
op = LinearRegressionIndices(with_geom=True)
gpd2 = STGeoProcess(op, gpd).run()

fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
gpd.boundary.plot(ax=ax, color="black")
gpd2.plot(ax=ax, color="red")
plt.show()
"""



'''
Created on 10 nov. 2023

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay
from shapely import MultiLineString, union_all
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class PrepareMasksLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def extrudeBuildingsOnFlatland(buildings, elevationFieldname="HAUTEUR", oriented=True,
                                   make_valid=True):
        result = buildings.copy(deep=True)
        if make_valid:
            result.geometry = result.geometry.apply(lambda g: g.buffer(0))
        result.geometry = result.apply(
            lambda rows: GeomLib.forceZCoordinateToZ0(
                rows.geometry, z0=rows[elevationFieldname]), axis=1)
        if oriented:
            result.geometry = result.geometry.apply(
                lambda g: GeomLib.reverseRingOrientation(g.normalize()))
        return result

    @staticmethod
    def getMasksAsBipoints(buildings, oriented=True, make_valid=True, union=False):
        # result = buildings.geometry.to_frame()
        result = buildings.copy(deep=True)
        if make_valid:
            result.geometry = result.geometry.apply(lambda g: g.buffer(0))
        if union:
            result = GeoDataFrame(
                [{"geometry": union_all(result.geometry)}], crs=buildings.crs)
        if oriented:
            result.geometry = result.geometry.apply(
                lambda g: GeomLib.reverseRingOrientation(g.normalize()))
        result.geometry = result.geometry.apply(lambda g: MultiLineString(
            GeomLib.toListOfBipointsAsLineStrings(g)))
        result = result.explode(ignore_index=True, index_parts=True)
        return result

    @staticmethod
    def removeNonVisible25DMasks(viewpoints, masks, elevationFieldname, horizon,
                                 h0, encode=False):
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(viewpoints, "GeoDataFrame")
        if not isinstance(masks, GeoDataFrame):
            raise IllegalArgumentTypeException(masks, "GeoDataFrame")
        if not elevationFieldname in masks:
            raise Exception(
                f"{elevationFieldname} is not a relevant field name!")
        if not GeoDataFrameLib.shareTheSameCrs(viewpoints, masks):
            raise Exception(
                "Illegal argument: viewpoints and masks are expected to share the same crs!")

        viewpoints2 = viewpoints.copy(deep=True)
        viewpoints2.geometry = viewpoints2.geometry.apply(
            lambda geom: geom if geom.has_z else GeomLib.forceZCoordinateToZ0(geom, h0))
        viewpoints2["__VIEWPOINT_GEOM__"] = viewpoints2.geometry
        viewpoints2["__VIEWPOINT_PK__"] = range(len(viewpoints2))
        viewpoints2.geometry = viewpoints2.geometry.apply(
            lambda g: g.buffer(horizon))

        masks2 = overlay(masks, viewpoints2, how="intersection")
        masks2.geometry = masks2.apply(
            lambda row: GeomLib.forceZCoordinateToZ0(
                row.geometry, row[elevationFieldname]-row.__VIEWPOINT_GEOM__.z),
            axis=1)
        masks2.geometry = masks2.geometry.apply(
            lambda g: GeomLib.reverseRingOrientation(g.normalize()))
        masks2.geometry = masks2.apply(
            lambda row: MultiLineString(
                GeomLib.toListOfBipointsAsLineStringsInFrontOf(
                    row.__VIEWPOINT_GEOM__, row.geometry)), axis=1)
        if encode:
            masks2.__VIEWPOINT_GEOM__ = masks2.__VIEWPOINT_GEOM__.apply(
                lambda geom: geom.wkt)
        return masks2

    @staticmethod
    def test(ofile=None):
        import matplotlib.pyplot as plt
        from shapely import Point
        from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
        # from t4gpd.morph.STGrid import STGrid

        # buildings = GeoDataFrameDemos.regularGridOfPlots2(nlines=4, ncols=4)
        # buildings["HAUTEUR"] = 10
        # vp = STGrid(buildings, dx=100, intoPoint=True, indoor=False).run()

        buildings = GeoDataFrameDemos.ensaNantesBuildings()

        viewpoints = GeoDataFrame(
            [{"gid": 100, "geometry": Point([355315, 6688413])}], crs=buildings.crs)

        masks = PrepareMasksLib.removeNonVisible25DMasks(
            viewpoints, buildings, elevationFieldname="HAUTEUR", horizon=40,
            h0=1.10, encode=True)

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
        buildings.plot(ax=ax, color="grey")
        viewpoints.plot(ax=ax, column="gid", marker="+")
        masks.plot(ax=ax, column="gid", linewidth=5)
        plt.tight_layout()
        plt.show()

        if not ofile is None:
            buildings.to_file(ofile, layer="buildings")
            viewpoints.to_file(ofile, layer="viewpoints")
            masks.to_file(ofile, layer="masks")

        return masks


# masks = PrepareMasksLib.test("/tmp/PrepareMasksLib.gpkg")



'''
Created on 12 sept. 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import Polygon
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.GeoProcess import GeoProcess


class AbstractHardShadow(GeoProcess):
    '''
    classdocs
    '''

    def run(self):
        rows = []

        for _dt, _radDir, _solarAlti, _solarAzim in self.sunPositions:
            # Hypothesis: the shadow cast by a 1m high mask is at most 10m long (arctan(0.1)=5.71)
            if (_solarAlti < AngleLib.toRadians(5.71)):
                print("Solar height less than 5.71 degrees (%.1f degrees) on %s!" % (
                    AngleLib.toDegrees(_solarAlti), _dt))
            else:
                for _, row in self.gdf.iterrows():
                    _shadow = self._auxiliary(
                        row, _radDir, _solarAlti, _solarAzim)
                    if _shadow is not None:
                        _row = self.updateOrAppend(
                            row, {"datetime": str(_dt), "geometry": _shadow})
                        rows.append(_row)

        if (0 == len(rows)):
            result = GeoDataFrame([{
                "datetime": dt,
                "geometry": Polygon([])
            } for dt, _, _, _ in self.sunPositions], crs=self.crs)
        else:
            result = GeoDataFrame(rows, crs=self.crs)

        if self.aggregate:
            result = result[["datetime", "geometry"]].dissolve(by="datetime")
            # Use a buffer to avoid slivers
            result.geometry = result.buffer(0.001, -1)
            result.reset_index(inplace=True)

        return result



'''
Created on 10 nov. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import overlay
from numpy import asarray, exp, zeros
from scipy.stats import kurtosis, skew
from shapely import LineString, Point, Polygon
from t4gpd.commons.Entropy import Entropy
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.raycasting.PanopticRaysLib import PanopticRaysLib


class RayCasting2DLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def get2DPanopticRaysGeoDataFrame(sensors, rayLength=100.0, nRays=64):
        return PanopticRaysLib.get2DGeoDataFrame(sensors, rayLength, nRays)

    @staticmethod
    def __keepTheAnchoredSegment(sensor, segments):
        epsilon = 1e-3
        if isinstance(segments, (Point, LineString)):
            if (epsilon > sensor.distance(segments)):
                return segments
        else:
            for segment in segments.geoms:
                if (epsilon > sensor.distance(segment)):
                    return segment
        return LineString([])

    @staticmethod
    def __buildIsovist(nRays, viewpoint, ray_ids, mls):
        if not GeomLib.isMultipart(mls):
            return Polygon([])

        _ctrPts1 = [ls.coords[-1] for ls in mls.geoms]
        if (nRays == len(_ctrPts1)):
            return Polygon(_ctrPts1)

        viewpoint = viewpoint.centroid.coords[0][0:2]
        assert len(ray_ids) == len(_ctrPts1)
        assert ray_ids == sorted(ray_ids)
        _ctrPts2 = []
        for i, ray_id in enumerate(ray_ids):
            if (0 == i):
                if (0 != ray_id):
                    _ctrPts2.append(viewpoint)
                _ctrPts2.append(_ctrPts1[i])
            else:
                if (ray_ids[i-1]+1 < ray_id):
                    _ctrPts2.append(viewpoint)
                _ctrPts2.append(_ctrPts1[i])
        return Polygon(_ctrPts2)

    @staticmethod
    def multipleRayCast2D(buildings, rays, withIndices, threshold=1e-3):
        if not GeoDataFrameLib.shareTheSameCrs(buildings, rays):
            raise Exception(
                "Illegal argument: buildings and rays must share shames CRS!")

        nRays = len(rays)
        rays = overlay(rays, buildings, how="difference", keep_geom_type=False)
        rays.geometry = rays.apply(lambda row: RayCasting2DLib.__keepTheAnchoredSegment(
            row.viewpoint, GeomLib.removeZCoordinate(row.geometry)), axis=1)
        rays = rays.loc[rays[~rays.geometry.apply(
            lambda geom: geom.is_empty)].index, :]
        # debug 10.11.2023
        rays = rays.loc[rays[rays.geometry.apply(
            lambda geom: geom.length > threshold)].index, :]
        isovRaysField = rays.dissolve(
            by="__VPT_ID__", as_index=False, aggfunc=list)
        isovRaysField.drop(columns=["__VPT_ID__"], inplace=True)
        for fieldname in isovRaysField.columns:
            if fieldname not in ["geometry", "__RAY_ID__"]:
                isovRaysField[fieldname] = isovRaysField[fieldname].apply(
                    lambda t: t[0])

        if (0 < len(isovRaysField)):
            isovRaysField["__ISOV_GEOM__"] = isovRaysField.apply(lambda row: RayCasting2DLib.__buildIsovist(
                nRays, row.viewpoint, row.__RAY_ID__, row.geometry), axis=1)

        if withIndices:
            isovRaysField = RayCasting2DLib.__addIndicesToIsovRaysField2D(
                nRays, isovRaysField)

        isovField = isovRaysField.copy(deep=True)

        # debug 20.08.2024
        isovField = isovField.set_geometry("__ISOV_GEOM__", crs=buildings.crs)
        isovField.drop(columns=["__RAY_ID__", "geometry"], inplace=True)
        isovField.rename_geometry("geometry", inplace=True)

        isovRaysField.drop(
            columns=["__RAY_ID__", "__ISOV_GEOM__"], inplace=True)

        return isovRaysField, isovField

    @staticmethod
    def __from2DRaysToRayLengths(nRays, ray_ids, mls):
        _raylens = asarray(
            [ray.length for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_raylens)):
            return _raylens

        raylens = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            raylens[ray_id] = _raylens[i]
        return raylens

    @staticmethod
    def __add2DIndices(isovRaysField, precision=1.0, base=exp(1)):
        isovRaysField["w_mean"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.mean()))
        isovRaysField["w_std"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.std()))
        isovRaysField["w_kurtosis"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(kurtosis(raylens, fisher=True, bias=True)))
        isovRaysField["w_skew"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(skew(raylens)))
        isovRaysField["w_entropy"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: Entropy.createFromDoubleValuesArray(
                raylens, precision).h(base)
        )
        return isovRaysField

    @staticmethod
    def __addIndicesToIsovRaysField2D(nRays, isovRaysField, precision=1.0, base=exp(1)):
        isovRaysField["__RAY_LEN__"] = isovRaysField.apply(
            lambda row: RayCasting2DLib.__from2DRaysToRayLengths(nRays, row.__RAY_ID__, row.geometry), axis=1)

        isovRaysField = RayCasting2DLib.__add2DIndices(
            isovRaysField, precision, base)

        isovRaysField["__ISOV_CENTRE__"] = isovRaysField.__ISOV_GEOM__.apply(
            lambda geom: geom.centroid)

        isovRaysField["vect_drift"] = isovRaysField.apply(
            lambda row: None if row.__ISOV_CENTRE__.is_empty else LineString([
                row.viewpoint.centroid.coords[0][0:2], row.__ISOV_CENTRE__.coords[0]]), axis=1)
        isovRaysField["drift"] = isovRaysField.vect_drift.apply(
            lambda v: None if v is None else v.length)

        isovRaysField.drop(
            columns=["__RAY_LEN__", "__ISOV_CENTRE__"], inplace=True)
        return isovRaysField



'''
Created on 5 feb. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arcsin, arctan2, asarray, cos, linspace, mean, pi, sin, sqrt
from shapely import Point, Polygon
from t4gpd.commons.CartesianProductLib import CartesianProductLib
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.proj.DoubleProjectionLib import DoubleProjectionLib


class DoubleProjectionOfSpheresLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def __tree_crown_mask_diameter(viewpoint, treeCrownCenter, treeCrownRadius):
        Ox, Oy, Oz = viewpoint.coords[0]
        Ax, Ay, Az = treeCrownCenter.coords[0]

        u = (Ax-Ox, Ay-Oy, 0)
        normU = sqrt(u[0]**2 + u[1]**2)
        u = (u[0]/normU, u[1]/normU, 0)

        r, d = treeCrownRadius, sqrt((Ax-Ox)**2 + (Ay-Oy)**2 + (Az-Oz)**2)

        if (d > r):
            beta = arcsin(r / d)
            if (Ax == Ox):
                alpha = pi / 2
            else:
                alpha = arctan2(Az-Oz, normU)

            magn = sqrt(d**2 - r**2)
            B = [
                Ox + cos(alpha + beta) * magn * u[0],
                Oy + cos(alpha + beta) * magn * u[1],
                Oz + sin(alpha + beta) * magn
            ]
            C = [
                Ox + cos(alpha - beta) * magn * u[0],
                Oy + cos(alpha - beta) * magn * u[1],
                Oz + sin(alpha - beta) * magn
            ]
            return [Ox, Oy, Oz], [Ax, Ay, Az], B, C
        raise Exception("TODO: To be developed!")

    @staticmethod
    def __tree_crown_mask_contour(O, A, B, npts):
        (Ox, Oy, Oz), (Ax, Ay, Az), (Bx, By, Bz) = O, A, B
        AA = asarray(A)
        u = asarray((Ax-Ox, Ay-Oy, Az-Oz))
        invNormU = 1 / sqrt(u[0]**2 + u[1]**2 + u[2]**2)
        u = invNormU * u
        v = asarray((Bx-Ax, By-Ay, Bz-Az))
        w = asarray(GeomLib3D.crossProduct(u, v))

        thetas = linspace(0, 2*pi, npts, endpoint=False)
        pts = [Point(AA + cos(theta) * v + sin(theta) * w) for theta in thetas]
        return pts

    @staticmethod
    def __projection_of_tree_crown(prj, viewpoint, treeCrownCenter,
                                   treeCrownRadius, size, npts):
        O, A, B, C = DoubleProjectionOfSpheresLib.__tree_crown_mask_diameter(
            viewpoint, treeCrownCenter, treeCrownRadius)

        if ("_stereographic_projection" == prj.__name__):
            ppB = prj(O, B, size)
            ppC = prj(O, C, size)

            depth = (ppB.z + ppC.z) / 2
            diameter = sqrt((ppC.x - ppB.x)**2 + (ppC.y - ppB.y)**2) / 2
            center = Point([(ppB.x + ppC.x) / 2, (ppB.y + ppC.y) / 2])
            disk = center.buffer(diameter/2)

            return GeomLib.forceZCoordinateToZ0(disk, z0=depth)

        pts = DoubleProjectionOfSpheresLib.__tree_crown_mask_contour(
            O, A, B, npts)
        ppts = [prj(viewpoint, p, size) for p in pts]
        z0 = mean([p.z for p in ppts])
        return GeomLib.forceZCoordinateToZ0(Polygon(ppts), z0=z0)

    @staticmethod
    def trees(sensors, trees, horizon, crownRadiusFieldname,
              h0=0.0, size=1, projectionName="Stereographic", npts=4):
        if not GeoDataFrameLib.shareTheSameCrs(trees, sensors):
            raise Exception(
                "Illegal argument: trees and sensors must share shames CRS!")

        prj = DoubleProjectionLib.projection_switch(projectionName)

        sensors2 = sensors.copy(deep=True)
        sensors2.geometry = sensors2.geometry.apply(
            lambda geom: geom if geom.has_z else GeomLib.forceZCoordinateToZ0(geom, h0))

        # PRE-PROCESSING
        result = CartesianProductLib.product_within_distance2(
            sensors2, trees, horizon)

        # PROCESSING
        result["geometry"] = result.apply(
            lambda row: DoubleProjectionOfSpheresLib.__projection_of_tree_crown(
                prj, row.geometry_x, row.geometry_y, row[crownRadiusFieldname], size=size, npts=npts),
            axis=1)

        # POST-PROCESSING
        result["depth"] = result.geometry.apply(
            lambda geom: geom.exterior.coords[0][2])

        return result

    def test():
        import matplotlib.pyplot as plt
        from matplotlib_scalebar.scalebar import ScaleBar
        from shapely.geometry import box
        from t4gpd.commons.proj.DoubleProjectionOfWallsLib import DoubleProjectionOfWallsLib
        from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
        from t4gpd.morph.STGrid import STGrid

        buildings = GeoDataFrameDemos.ensaNantesBuildings()
        trees = GeoDataFrameDemos.ensaNantesTrees()

        trees.geometry = trees.geometry.apply(
            lambda geom: GeomLib.forceZCoordinateToZ0(geom, 12))
        trees["crown_radius"] = 4

        dx = 100
        grid = STGrid(buildings, dx=dx, dy=None, indoor=False, intoPoint=True,
                      encode=True, withDist=False).execute()  # < 15 sec
        sensors = grid

        size = 4
        # projectionName = "Isoaire"
        # projectionName = "orthogonal"
        projectionName = "Stereographic"
        pbuildings = DoubleProjectionOfWallsLib.walls(
            sensors, buildings, horizon=50.0, elevationFieldname="HAUTEUR",
            h0=0.0, size=size, projectionName=projectionName, npts=5, aggregate=False)
        ptrees = DoubleProjectionOfSpheresLib.trees(
            sensors, trees, horizon=50.0, crownRadiusFieldname="crown_radius",
            h0=0.0, size=size, projectionName=projectionName)

        # minx, miny, maxx, maxy = box(
        #     *sensors[sensors.gid == 6].total_bounds).buffer(6.0).bounds
        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.35 * 8.26))
        ax.set_title(f"Nantes (44) ({projectionName})", size=28)
        buildings.plot(ax=ax, color="grey")
        sensors.plot(ax=ax, color="black", marker="P")
        trees.plot(ax=ax, color="green", marker=".")

        pbuildings.plot(ax=ax, column="depth", cmap="Spectral",
                        edgecolor="black", alpha=0.75)
        ptrees.plot(ax=ax, column="depth", cmap="Spectral",
                    edgecolor="black", alpha=0.75, legend="True")

        ax.axis("off")
        # ax.axis([minx, maxx, miny, maxy])
        # ax.legend(loc="lower left", fontsize=18)
        scalebar = ScaleBar(dx=1.0, units="m", length_fraction=None, box_alpha=0.85,
                            width_fraction=0.005, location="lower right", frameon=True)
        ax.add_artist(scalebar)
        fig.tight_layout()
        plt.show()
        plt.close(fig)

        return ptrees


# ptrees = DoubleProjectionOfSpheresLib.test()



'''
Created on 11 juin 2020

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.grid.GridLib import GridLib


class STGrid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, dy=None, indoor=None, intoPoint=True,
                 encode=True, withDist=False, roi=None):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        self.gdf = gdf
        self.dx = dx
        self.dy = dy
        if indoor in [None, True, False, "both"]:
            self.indoor = indoor
        else:
            raise Exception(
                "Illegal argument: indoor must be chosen in [None, True, False, 'both']!")
        self.intoPoint = intoPoint
        self.encode = encode
        self.withDist = withDist
        self.roi = roi

    def run(self):
        if self.roi is None:
            grid = GridLib.getGrid2(self.gdf, self.dx, self.dy, self.encode)
        else:
            grid = GridLib.getGrid2(self.roi, self.dx, self.dy, self.encode)

        if self.indoor is None:
            pass
        elif ("both" == self.indoor):
            grid = GridLib.getIndoorOutdoorGrid(self.gdf, grid)
        elif self.indoor:
            grid = GridLib.getIndoorGrid(self.gdf, grid)
        else:
            grid = GridLib.getOutdoorGrid(self.gdf, grid)

        if self.intoPoint:
            grid.geometry = grid.centroid

        if self.withDist:
            if (0 == len(self.gdf)):
                grid["dist_to_ctr"] = float("inf")
            else:
                grid = GridLib.getDistanceToNearestContour(self.gdf, grid)

        return grid


"""
# -----
import matplotlib.pyplot as plt
from geopandas import clip
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.demos.NantesBDT import NantesBDT
from t4gpd.morph.STBBox import STBBox

buffDist = 100
roi = STBBox(GeoDataFrameDemos.squaresInNantes("Royale"), buffDist).run()
buildings = NantesBDT.buildings(roi)

dx = 30
sensors = STGrid(buildings, dx=dx, dy=None, indoor=False,
	intoPoint=False, withDist=True).execute()

# -----
fig, ax = plt.subplots(figsize=(1*8.26, 1*8.26))
buildings.plot(ax=ax, color="grey")
sensors.boundary.plot(ax=ax, color="black", linewidth=0.15)
fig.tight_layout()
plt.show()
plt.close(fig)
"""



'''
Created on 12 feb. 2021

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay
from pandas import concat
from t4gpd.commons.ConcaveLib import ConcaveLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPolygonize import STPolygonize


class STMakeBlocks(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, roads, roi=None):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, "GeoDataFrame")
        self.buildings = buildings

        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, "GeoDataFrame")
        if not ((roi is None) or isinstance(roi, GeoDataFrame)):
            raise IllegalArgumentTypeException(roi, "GeoDataFrame or None")
        self.linestrings = STMakeBlocks.__prepare_linestrings(roads, roi)

    @staticmethod
    def __prepare_linestrings(roads, roi):
        if roi is None:
            return roads
        roi.geometry = roi.geometry.boundary
        return GeoDataFrame(concat([roads[["geometry"]], roi[["geometry"]]]), crs=roads.crs)

    def __mkBlocks(self):
        # POLYGONIZE LINESTRINGS
        polygons = STPolygonize(self.linestrings).run()

        # GROUP BUILDINGS PER URBAN BLOCKS
        buildings = overlay(
            self.buildings[["geometry"]], polygons, how="intersection")
        blocks = buildings.dissolve(by="gid", as_index=False)

        # GROUP LINESTRINGS PER URBAN BLOCKS
        linestrings = overlay(
            self.linestrings, polygons, how="intersection", keep_geom_type=True)
        linestrings = linestrings.dissolve(by="gid", as_index=False)

        # TRY TO REMOVE CONCAVITIES IN EACH BLOCK
        blocks.geometry = blocks.apply(
            lambda row: ConcaveLib.fillInTheConcavities(
                row.geometry, linestrings.at[row.gid, "geometry"]), axis=1)

        return blocks

    def run(self):
        return self.__mkBlocks()



'''
Created on 10 mar. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timedelta
from ephem import (localtime, next_first_quarter_moon, next_full_moon, next_last_quarter_moon, next_new_moon,
                   previous_first_quarter_moon, previous_full_moon, previous_last_quarter_moon, previous_new_moon)
from pandas import DataFrame


class MoonLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def moon_phases_calendar(year):
        # Inner function
        def _add(year, moons, dt, fct1, fct2, label):
            for fct in [fct1, fct2]:
                t = localtime(fct(dt))
                if (year == t.year):
                    moons[t.date()] = label

        # 1st step
        moons = {}
        dt0 = datetime(year, 1, 1, 12)
        for day_of_year in range(0, 366, 7):
            dt = dt0 + timedelta(days=day_of_year)
            _add(year, moons, dt, next_first_quarter_moon,
                 previous_first_quarter_moon, "first_quarter")
            _add(year, moons, dt, next_full_moon, previous_full_moon, "full")
            _add(year, moons, dt, next_last_quarter_moon,
                 previous_last_quarter_moon, "last_quarter")
            _add(year, moons, dt, next_new_moon, previous_new_moon, "new")

        # 2nd step
        rows = []
        for day in sorted(moons.keys()):
            rows.append({
                "timestamp": day,
                "day_of_year": day.timetuple().tm_yday,
                "moon_phasis": moons[day],
            })
        df = DataFrame(rows)
        return df


# df = MoonLib.moon_phases_calendar(2023)



'''
Created on 30 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.CartesianProductLib import CartesianProductLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.proj.DoubleProjectionLib import DoubleProjectionLib


class DoubleProjectionOfPointsLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def points(sensors, streetlights, horizon=None, h0=0.0, size=1, projectionName="Stereographic",
               encode=True):
        if not GeoDataFrameLib.shareTheSameCrs(sensors, streetlights):
            raise Exception(
                "Illegal argument: sensors and streetlights must share shames CRS!")

        prj = DoubleProjectionLib.projection_switch(projectionName)

        sensors2 = sensors.copy(deep=True)
        sensors2.geometry = sensors2.geometry.apply(
            lambda geom: geom if geom.has_z else GeomLib.forceZCoordinateToZ0(geom, h0))

        if horizon is None:
            result = CartesianProductLib.product(sensors2, streetlights)
        else:
            result = CartesianProductLib.product_within_distance2(
                sensors2, streetlights, horizon)

        result["geometry"] = result.apply(
            lambda row: prj(row.geometry_x, row.geometry_y, size=size), axis=1)
        result["depth"] = result.geometry.apply(lambda p: p.z)
        result.geometry = result.apply(
            lambda row: GeomLib.forceZCoordinateToZ0(
                row.geometry, row.geometry_x.z), axis=1)

        if encode:
            result.geometry_x = result.geometry_x.apply(lambda g: g.wkt)
            result.geometry_y = result.geometry_y.apply(lambda g: g.wkt)

        result.rename(columns={
            "geometry_x": "observer_p",
            "geometry_y": "remote_pos",
        }, inplace=True)

        return GeoDataFrame(result, geometry="geometry", crs=sensors.crs)

    @staticmethod
    def test():
        import matplotlib.pyplot as plt
        from matplotlib_scalebar.scalebar import ScaleBar
        from geopandas import clip
        from t4gpd.demos.GeoDataFrameDemosC import GeoDataFrameDemosC
        from t4gpd.morph.STGrid import STGrid

        iris = GeoDataFrameDemosC.irisTastavin()
        buildings = GeoDataFrameDemosC.irisTastavinBuildings()
        roads = GeoDataFrameDemosC.irisTastavinRoads()
        streetlights = GeoDataFrameDemosC.irisTastavinStreetlights()
        trees = GeoDataFrameDemosC.irisTastavinTrees()

        streetlights.geometry = streetlights.geometry.apply(
            lambda geom: GeomLib.forceZCoordinateToZ0(geom, 12))

        dx = 150
        grid = STGrid(buildings, dx=dx, dy=None, indoor=False, intoPoint=True,
                      encode=True, withDist=False).execute()  # < 15 sec
        sensors = clip(grid, iris, keep_geom_type=True)

        size = 4
        projectionName = "Isoaire"
        projectionName = "Orthogonal"
        projectionName = "Stereographic"
        pp = DoubleProjectionOfPointsLib.points(
            sensors, streetlights, horizon=100, h0=0.0,
            size=size, projectionName=projectionName, encode=False)

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.35 * 8.26))
        ax.set_title(
            f"Montpellier (34), IRIS Tastavin ({projectionName})", size=28)
        iris.boundary.plot(ax=ax, color="red", linestyle="dotted")
        buildings.plot(ax=ax, color="grey")
        # big_buildings.plot(ax=ax, color="lightgrey")
        # grid.boundary.plot(ax=ax, color="green")
        sensors.buffer(size).boundary.plot(ax=ax, color="red")
        pp.plot(ax=ax, column="depth", cmap="Spectral",
                marker="+", legend=True)
        # skymaps.plot(ax=ax, color="black", linewidth=0.15)
        ax.axis("off")
        # ax.legend(loc="lower left", fontsize=18)
        scalebar = ScaleBar(dx=1.0, units="m", length_fraction=None, box_alpha=0.85,
                            width_fraction=0.005, location="lower right", frameon=True)
        ax.add_artist(scalebar)
        fig.tight_layout()
        plt.show()
        plt.close(fig)

        return pp


# pp = DoubleProjectionOfPointsLib.test()



'''
Created on 24 jul. 2023

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import arctan2, asarray, cos, full, linspace, pi, sqrt, sin
from shapely import Polygon
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.raycasting.RayCasting25DLib import RayCasting25DLib


class STSkyMap25D(GeoProcess):
    '''
    classdocs
    '''
    PIDIV2 = 0.5 * pi

    def __init__(self, buildings, viewpoints, nRays=64, rayLength=100.0,
                 elevationFieldname="HAUTEUR", h0=0.0, size=4.0, epsilon=1e-2,
                 projectionName="Stereographic", withIndices=False, withAngles=False,
                 encode=False, threshold=1e-6):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(
                buildings, "buildings GeoDataFrame")
        if not elevationFieldname in buildings:
            raise Exception(
                f"{elevationFieldname} is not a relevant field name!")
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(
                viewpoints, "viewpoints GeoDataFrame")
        if not "gid" in viewpoints:
            raise Exception(
                "viewpoints must have a 'gid' field name (with unique values)!")

        if not GeoDataFrameLib.shareTheSameCrs(buildings, viewpoints):
            raise Exception(
                "Illegal argument: buildings and viewpoints must share shames CRS!")

        self.viewpoints = viewpoints
        self.viewpoints.geometry = self.viewpoints.geometry.apply(
            lambda geom: GeomLib.forceZCoordinateToZ0(
                geom, z0=GeomLib3D.centroid(geom).z if geom.has_z else h0))

        self.buildings = buildings
        self.elevationFieldname = elevationFieldname

        # CHECK IF buildings IS NOT EMPTY
        if (0 < len(self.buildings)):
            # CHECK IF ANY elevationFieldname VALUE IS NULL OR NaN
            if (any(self.buildings[self.elevationFieldname].isna()) or
                    any(self.buildings[self.elevationFieldname].isnull())):
                raise Exception(
                    "There is at least one NaN value in the elevation column of the buildings dataframe!")

            # CLEAN GEOMETRIES
            self.buildings.geometry = self.buildings.geometry.apply(
                lambda g: g.buffer(0))
            self.buildings.geometry = self.buildings.apply(
                lambda row: GeomLib.forceZCoordinateToZ0(
                    row.geometry, row[self.elevationFieldname]),
                axis=1
            )

        self.nRays = nRays
        self.rayLength = rayLength
        self.rays = RayCasting25DLib.get25DPanopticRaysGeoDataFrame(
            self.viewpoints, rayLength, nRays, h0)

        self.size = size
        self.epsilon = epsilon
        if not projectionName in ["Isoaire", "Orthogonal", "Stereographic"]:
            raise IllegalArgumentTypeException(
                projectionName, "spherical projection as 'Stereographic', 'Orthogonal', 'Isoaire'")
        if ("Isoaire" == projectionName):
            self.proj = self.__isoaire
        elif ("Orthogonal" == projectionName):
            self.proj = self.__orthogonal
        elif ("Stereographic" == projectionName):
            self.proj = self.__stereographic

        self.withIndices = withIndices
        self.withAngles = withAngles
        self.encode = encode
        self.threshold = threshold

    def __isoaire(self, lat, lon):
        radius = sqrt((1 - sin(lat)) / ((cos(lat)*cos(lon))
                      ** 2 + (cos(lat)*sin(lon))**2))
        radius *= (self.size * cos(lat))
        return (radius * cos(lon), radius * sin(lon))

    def __orthogonal(self, lat, lon):
        radius = (self.size * cos(lat))
        return (radius * cos(lon), radius * sin(lon))

    def __stereographic(self, lat, lon):
        radius = 1.0 / (1.0 + sin(lat))
        radius *= (self.size * cos(lat))
        return (radius * cos(lon), radius * sin(lon))

    def __angles(self, heights, widths):
        return [arctan2(h, w) for h, w in zip(heights, widths)]

    def __buildSkyMap(self, viewpoint, lats, lons):
        viewpoint = viewpoint.centroid
        pnodes = [self.proj(lat, lon) for lat, lon in zip(lats, lons)]
        pnodes = [(viewpoint.x + pp[0], viewpoint.y + pp[1]) for pp in pnodes]
        return Polygon(viewpoint.buffer(self.size + self.epsilon).exterior.coords, [pnodes])

    def run(self):
        smapRaysField = RayCasting25DLib.multipleRayCast25D(
            self.viewpoints, self.buildings, self.rays, self.nRays,
            self.elevationFieldname, self.withIndices, h0=0.0, threshold=self.threshold)
        # smapRaysField.to_csv("/tmp/7.csv") # DEBUG

        if (0 < len(smapRaysField)):
            smapRaysField["angles"] = smapRaysField.apply(
                lambda row: self.__angles(
                    row.__RAY_DELTA_ALT__, row.__RAY_LEN__),
                axis=1
            )
            # smapRaysField.to_csv("/tmp/8.csv") # DEBUG
            lons = linspace(0, 2*pi, self.nRays, endpoint=False)
            smapRaysField.geometry = smapRaysField.apply(lambda row: self.__buildSkyMap(
                row.viewpoint, row.angles, lons), axis=1)
            smapRaysField.angles = smapRaysField.angles.apply(
                lambda v: (180/pi) * asarray(v))
            # smapRaysField.to_csv("/tmp/9.csv") # DEBUG

        fields = ["__RAY_ID__"]
        for field in ["__RAY_LEN__", "__RAY_ALT__", "__RAY_DELTA_ALT__", "__HAUTEUR_VP__"]:
            if field in smapRaysField.columns:
                fields.append(field)

        if not self.withAngles:
            fields += ["angles"]

        smapRaysField.drop(columns=fields, inplace=True)

        if self.encode:
            smapRaysField.viewpoint = smapRaysField.viewpoint.apply(
                lambda vp: vp.wkt)
            if "angles" in smapRaysField:
                smapRaysField.angles = smapRaysField.angles.apply(
                    lambda a: ArrayCoding.encode(a))

        return smapRaysField


"""
import matplotlib.pyplot as plt
from shapely import Point
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos

buildings = GeoDataFrameDemos.districtRoyaleInNantesBuildings()
pts = [ Point((355119.8, 6689339.2)), Point((355143.0, 6689359.4)),
	Point((355113.4, 6689397.9)) ]
sensors = GeoDataFrame([{"gid": i, "geometry": p} for i, p in enumerate(pts)],
	crs=buildings.crs)
# sensors.geometry = sensors.geometry.apply(lambda g: g.buffer(5))

projectionName = "Stereographic"
projectionName = "Isoaire"
# projectionName = "Orthogonal"
skymaps = STSkyMap25D(buildings, sensors, projectionName=projectionName,
                      withIndices=True).run()

print(skymaps)
minx, miny, maxx, maxy = skymaps.total_bounds
fig, ax = plt.subplots(figsize=(8.26, 8.26))
ax.set_title(projectionName, fontsize=20)
buildings.plot(ax=ax, color="grey")
sensors.plot(ax=ax, color="red")
skymaps.plot(ax=ax, color="blue")
ax.axis([minx, maxx, miny, maxy])
ax.axis("off")
fig.tight_layout()
plt.show()
"""

"""
from shapely import Point

h = 10.0
masks = GeoDataFrame([{
    "gid": 1,
    "geometry": Polygon([(0, 0), (0, 10), (10, 10), (10, 9), (1, 9), (1, 0), (0, 0)]),
    "HAUTEUR": h
}])
sensors = GeoDataFrame([
    {"gid": 1, "geometry": Point([1.1, 8.9])},
])
result = STSkyMap25D(masks, sensors,
                     nRays=64, rayLength=100.0,
                     elevationFieldname="HAUTEUR",
                     h0=0.0, size=2.0, epsilon=1e-2,
                     projectionName="Stereographic",
                     withIndices=True,
                     withAngles=False).run()
print(result[['gid', 'viewpoint', 'w_mean', 'w_std', 'h_mean', 'h_over_w',
       'svf']])
"""

"""
from io import StringIO
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.demos.NantesBDT import NantesBDT
from t4gpd.morph.STBBox import STBBox

_sio = StringIO('''gid;geometry
150755;"POLYGON ((355050.15 6689030.25, 355050.15 6689032.25, 355048.15 6689032.25, 355048.15 6689030.25, 355050.15 6689030.25))"
''')
sensors = GeoDataFrameLib.read_csv(_sio)
# buildings = NantesBDT.buildings(STBBox(sensors, 10).run())
buildings = NantesBDT.buildings(STBBox(sensors, 200).run())
buildings = buildings.loc[buildings[buildings.ID.isin(
    ["BATIMENT0000000302927659", "BATIMENT0000000302930463",
     "BATIMENT0000000302930192"])].index]
sensors.to_file("/tmp/sensors.shp")
buildings.to_file("/tmp/buildings.shp")
skymaps = STSkyMap25D(buildings, sensors, nRays=32, rayLength=200,
                      elevationFieldname="HAUTEUR", h0=0.0, size=7.0,
                      withIndices=True, withAngles=True, encode=True,
                      threshold=1e-7).run()
# skymaps.to_file("/tmp/skymaps.shp")
"""

"""
from io import StringIO
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.demos.NantesBDT import NantesBDT

_sio = StringIO('''gid;geometry
86751;"POLYGON ((355042.15 6688780.25, 355042.15 6688782.25, 355040.15 6688782.25, 355040.15 6688780.25, 355042.15 6688780.25))"
''')
sensors = GeoDataFrameLib.read_csv(_sio)
buildings = NantesBDT.buildings()
buildings = buildings.loc[buildings[buildings.ID == "BATIMENT0000000302923050"].index]
sensors.to_file("/tmp/sensors.shp")
buildings.to_file("/tmp/buildings.shp")
skymaps = STSkyMap25D(buildings, sensors, nRays=8, rayLength=200,
                      elevationFieldname="HAUTEUR", h0=0.0, size=7.0,
                      withIndices=True, withAngles=True, encode=True,
                      threshold=1e-7).run()
# skymaps.to_file("/tmp/skymaps.shp")
"""

'''
from io import StringIO

_sio = StringIO("""gid,row,column,dist_to_ctr,dalle_id,geometry
0,0,0,44.89237128934325,488,"POLYGON ((354596.15 6688442.25, 354596.15 6688444.25, 354594.15 6688444.25, 354594.15 6688442.25, 354596.15 6688442.25))"
""")
sensors = GeoDataFrameLib.read_csv(_sio, sep=",")

_sio = StringIO("""ID,NATURE,USAGE1,USAGE2,LEGER,ETAT,DATE_CREAT,DATE_MAJ,DATE_APP,DATE_CONF,SOURCE,ID_SOURCE,ACQU_PLANI,PREC_PLANI,ACQU_ALTI,PREC_ALTI,NB_LOGTS,NB_ETAGES,MAT_MURS,MAT_TOITS,HAUTEUR,Z_MIN_SOL,Z_MIN_TOIT,Z_MAX_TOIT,Z_MAX_SOL,ORIGIN_BAT,APP_FF,geometry
BATIMENT0000000302923133,Indifférenciée,Indifférencié,,Non,En service,2012-06-19 18:43:00,2018-09-18 13:36:26,,,,,BDParcellaire recalée,3.0,Corrélation,1.0,,,,,6.2,6.0,12.2,16.0,6.2,Cadastre,,"POLYGON Z ((354657.1 6688439.9 12.2, 354664.3 6688441.8 12.2, 354668.8 6688424 12.2, 354661.7 6688422.2 12.2, 354657.1 6688439.9 12.2))"
BATIMENT0000000302923132,"Industriel, agricole ou commercial",Commercial et services,,Non,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,2011-01-01,,,,BDParcellaire recalée,3.0,Corrélation,1.0,0.0,1.0,,,13.4,6.0,19.5,24.2,6.5,Cadastre,C 0.4,"POLYGON Z ((354685.5 6688242.4 19.5, 354678.8 6688271.8 19.5, 354645 6688417.2 19.5, 354691.9 6688429.2 19.5, 354695.3 6688414.5 19.5, 354701.4 6688415.9 19.5, 354702.9 6688409.7 19.5, 354704.3 6688410 19.5, 354705.9 6688403.4 19.5, 354707.6 6688403.8 19.5, 354710 6688393.9 19.5, 354713.3 6688394.7 19.5, 354714.8 6688388.6 19.5, 354716.3 6688388.9 19.5, 354717.8 6688382.7 19.5, 354719.3 6688383 19.5, 354720.8 6688376.8 19.5, 354722.2 6688377.1 19.5, 354723.8 6688370.6 19.5, 354725.3 6688370.9 19.5, 354727 6688364.2 19.5, 354717.8 6688361.9 19.5, 354742.7 6688255.8 19.5, 354685.5 6688242.4 19.5))"
BATIMENT0000000302923159,Indifférenciée,Indifférencié,,Non,En service,2012-06-19 18:43:00,2018-09-18 13:36:27,,,,,BDParcellaire recalée,3.0,Interpolation bâti BDTopo,2.5,,,,,11.7,6.0,17.7,24.3,6.6,Cadastre,,"POLYGON Z ((354543.4 6688344.2 17.7, 354534.3 6688360.8 17.7, 354617.6 6688407.1 17.7, 354626.8 6688390.2 17.7, 354543.4 6688344.2 17.7))"
BATIMENT0000002335125943,Indifférenciée,Indifférencié,,Non,En service,2023-10-16 08:40:22,,,,DGFiP,,BDParcellaire,5.0,Pas de Z,9999.0,,,,,,,,,,Cadastre,,"POLYGON Z ((354627.8 6688496.5 -1000, 354637.3 6688498.8 -1000, 354643.7 6688473.3 -1000, 354634.2 6688470.9 -1000, 354627.8 6688496.5 -1000))"
BATIMENT0000002335125944,Indifférenciée,Indifférencié,,Oui,En service,2023-10-16 08:40:22,,,,DGFiP,,BDParcellaire,5.0,Pas de Z,9999.0,,,,,,,,,,Cadastre,,"POLYGON Z ((354626.4 6688497.1 -1000, 354627.8 6688496.5 -1000, 354634.2 6688470.9 -1000, 354643.7 6688473.3 -1000, 354645 6688472.7 -1000, 354633.2 6688469.6 -1000, 354626.4 6688497.1 -1000))"
BATIMENT0000002335125945,Indifférenciée,Indifférencié,,Oui,En service,2023-10-16 08:40:22,,,,DGFiP,,BDParcellaire,5.0,Pas de Z,9999.0,,,,,,,,,,Cadastre,,"POLYGON Z ((354627.8 6688496.5 -1000, 354626.4 6688497.1 -1000, 354638.2 6688500 -1000, 354645 6688472.7 -1000, 354643.7 6688473.3 -1000, 354637.3 6688498.8 -1000, 354627.8 6688496.5 -1000))"
BATIMENT0000000302923157,"Industriel, agricole ou commercial",Industriel,,Non,En service,2012-06-19 18:43:00,2018-09-18 13:36:27,,,,,BDParcellaire recalée,3.0,Interpolation bâti BDTopo,2.5,,,,,6.6,6.0,12.6,13.8,6.8,Cadastre,,"POLYGON Z ((354697.2 6688468.6 12.6, 354679.9 6688499.1 12.6, 354699.5 6688510.2 12.6, 354712.5 6688486.8 12.6, 354716.3 6688479.6 12.6, 354717.1 6688478.4 12.6, 354712.7 6688475.8 12.6, 354708.8 6688473.3 12.6, 354707.9 6688474.7 12.6, 354697.2 6688468.6 12.6))"
BATIMENT0000000302923139,"Industriel, agricole ou commercial",Commercial et services,,Oui,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,,,,,BDParcellaire recalée,3.0,Corrélation,1.0,0.0,1.0,,,7.9,6.3,14.2,16.6,6.6,Cadastre,B 1.0,"POLYGON Z ((354676.8 6688507.4 14.2, 354669.8 6688519.7 14.2, 354677.9 6688524.4 14.2, 354683.4 6688527.7 14.2, 354685.2 6688528.7 14.2, 354689.3 6688531 14.2, 354712.5 6688543.7 14.2, 354719.9 6688530.4 14.2, 354689.1 6688514 14.2, 354676.8 6688507.4 14.2))"
BATIMENT0000000302923135,Indifférenciée,Résidentiel,Commercial et services,Non,En service,2012-06-19 18:43:00,2023-01-31 17:40:48,1900-01-01,,,,BDParcellaire recalée,3.0,Corrélation,1.0,5.0,4.0,10,20,10.5,6.2,16.7,20.0,6.5,Cadastre,A 1.0,"POLYGON Z ((354677.9 6688524.4 16.7, 354669.8 6688519.7 16.7, 354664.6 6688529 16.7, 354676.4 6688535.4 16.7, 354680.2 6688528.6 16.7, 354676.5 6688526.6 16.7, 354677.9 6688524.4 16.7))"
BATIMENT0000000302923137,"Industriel, agricole ou commercial",Industriel,,Oui,En service,2012-06-19 18:43:00,2023-01-31 17:40:49,,,,,BDParcellaire recalée,3.0,Interpolation bâti BDTopo,2.5,,,,,4.9,6.0,10.9,15.4,6.6,Cadastre,,"POLYGON Z ((354680.5 6688537.6 10.9, 354684.6 6688539.7 10.9, 354686.6 6688535.9 10.9, 354689.3 6688531 10.9, 354685.2 6688528.7 10.9, 354683.4 6688527.7 10.9, 354678.8 6688536.8 10.9, 354680.5 6688537.6 10.9))"
BATIMENT0000000302923136,Indifférenciée,Commercial et services,Résidentiel,Non,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,1900-01-01,,,,BDParcellaire recalée,3.0,Interpolation bâti BDTopo,2.5,6.0,3.0,,,7.3,6.2,13.5,18.3,6.6,Cadastre,A 1.0,"POLYGON Z ((354664.6 6688529 13.5, 354659.1 6688538.6 13.5, 354660.4 6688542 13.5, 354671.5 6688548 13.5, 354678 6688536.3 13.5, 354676.4 6688535.4 13.5, 354664.6 6688529 13.5))"
BATIMENT0000000302923134,Indifférenciée,Résidentiel,Commercial et services,Non,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,1880-01-01,,,,BDParcellaire recalée,3.0,Corrélation,1.0,19.0,5.0,20,20,12.2,6.3,18.5,22.0,6.5,Cadastre,A 1.0,"POLYGON Z ((354684.6 6688539.7 18.5, 354680.5 6688537.6 18.5, 354678.8 6688536.8 18.5, 354678 6688536.3 18.5, 354671.5 6688548 18.5, 354683.8 6688554.7 18.5, 354689.5 6688545.1 18.5, 354683.3 6688542.2 18.5, 354684.6 6688539.7 18.5))"
BATIMENT0000000302923570,"Industriel, agricole ou commercial",Commercial et services,,Non,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,2007-01-01,,,,BDParcellaire recalée,3.0,Corrélation,1.0,0.0,1.0,,,9.0,6.4,15.5,18.5,6.6,Cadastre,C 0.4,"POLYGON Z ((354686.6 6688535.9 15.5, 354692.5 6688539.2 15.5, 354690.5 6688542.8 15.5, 354707.6 6688552.4 15.5, 354712.5 6688543.7 15.5, 354689.3 6688531 15.5, 354686.6 6688535.9 15.5))"
BATIMENT0000000302923571,Indifférenciée,Résidentiel,Commercial et services,Non,En service,2012-06-19 18:43:00,2019-03-15 08:35:34,2007-01-01,,,,BDParcellaire recalée,3.0,Corrélation,1.0,8.0,4.0,00,00,9.4,6.3,15.8,21.1,6.6,Cadastre,C 0.5,"POLYGON Z ((354683.8 6688554.7 15.8, 354692.6 6688559.5 15.8, 354698.9 6688562.9 15.8, 354701.9 6688562.4 15.8, 354707.6 6688552.4 15.8, 354690.5 6688542.8 15.8, 354689.5 6688545.1 15.8, 354683.8 6688554.7 15.8))"
""")
masks = GeoDataFrameLib.read_csv(_sio, sep=",")

masks.fillna({"HAUTEUR": 3.33}, inplace=True)
result = STSkyMap25D(masks, sensors, nRays=64, rayLength=100,
                     elevationFieldname="HAUTEUR", h0=0.0, size=1.0,
                     withIndices=True, withAngles=True, encode=True,
                     threshold=1e-7).run()
print(sensors.columns)
print(result.columns)
'''



'''
Created on 21 juin 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STMakeNeighborhood(object):
    '''
    classdocs
    '''

    def __init__(self, gdf, gidFieldName="gid"):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        if gidFieldName not in gdf:
            raise Exception(f"{gidFieldName} is not a relevant field name!")

        self.gdf = gdf
        self.gid = gidFieldName

    def run(self):
        def myAppend(list1, list2):
            if isinstance(list1, list) and isinstance(list2, list):
                return list1 + list2
            if isinstance(list1, list):
                return list1
            if isinstance(list2, list):
                return list2
            raise Exception("Unreachable instruction!")

        left, right = f"{self.gid}_left", f"{self.gid}_right"

        gdf1 = self.gdf.sjoin(self.gdf, how="inner", predicate="touches")
        gdf1 = gdf1.loc[:, [left, right]]
        gdf1 = gdf1.groupby(by=left, as_index=False).agg({right: list})

        gdf2 = self.gdf.sjoin(self.gdf, how="inner", predicate="overlaps")
        gdf2 = gdf2.loc[:, [left, right]]
        gdf2 = gdf2.groupby(by=left, as_index=False).agg({right: list})

        right1, right2 = f"{right}_x", f"{right}_y"
        gdf12 = gdf1.merge(gdf2, how="outer", on=left)
        gdf12["neighbors"] = gdf12.apply(
            lambda row: myAppend(row[right1], row[right2]), axis=1)
        gdf12 = gdf12.loc[:, [left, "neighbors"]]

        gdf = self.gdf.merge(
            gdf12, how="left", left_on=self.gid, right_on=left)
        gdf.drop(columns=[left], inplace=True)

        return gdf


"""
def _plot(gdf, neigh):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
    gdf.boundary.plot(ax=ax, color="grey", label="Input geoms")
    gdf.apply(lambda x: ax.annotate(
                text=x.gid, xy=x.geometry.centroid.coords[0],
                color="black", size=12  , ha="center", va="bottom"), axis=1)
    # neigh.boundary.plot(ax=ax, color="red", label="Result")
    neigh.apply(lambda x: ax.annotate(
                text=x.neighbors, xy=x.geometry.centroid.coords[0],
                color="red", size=12, ha="center", va="top"), axis=1)
    ax.legend(fontsize=14)
    fig.tight_layout()
    plt.show()
    plt.close(fig)
    
from shapely import box
from t4gpd.morph.STGrid import STGrid

gdf = GeoDataFrame([
    {"gid": 1, "geometry": box(0,0,3,3)},
])
grid = STGrid(gdf, dx=1, intoPoint=False).run()
# grid = grid.loc[:,["gid", "geometry"]]
neigh = STMakeNeighborhood(grid, "gid").run()

_plot(grid, neigh)
"""

"""
def _plot(gdf, neigh):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
    gdf.boundary.plot(ax=ax, color="grey", label="Input geoms")
    gdf.apply(lambda x: ax.annotate(
                text=x.gid, xy=x.geometry.centroid.coords[0],
                color="black", size=12  , ha="center", va="bottom"), axis=1)
    # neigh.boundary.plot(ax=ax, color="red", label="Result")
    neigh.apply(lambda x: ax.annotate(
                text=x.neighbors, xy=x.geometry.centroid.coords[0],
                color="red", size=12, ha="center", va="top"), axis=1)
    ax.legend(fontsize=14)
    fig.tight_layout()
    plt.show()
    plt.close(fig)

from geopandas import GeoDataFrame
from numpy.random import default_rng
from shapely.geometry import CAP_STYLE, Point
from t4gpd.morph.STVoronoiPartition2 import STVoronoiPartition2

npts, rng = 20, default_rng(1234)
XY = rng.uniform(low=0, high=100, size=(npts,2))
gdf = GeoDataFrame([{"gid": gid, "geometry": Point(xy)} for gid, xy in enumerate(XY)])
gdf.geometry = gdf.geometry.apply(lambda g: g.buffer(1, cap_style=CAP_STYLE.square))

plots = STVoronoiPartition2(gdf, max_segment_length=0.2).run()
neigh = STMakeNeighborhood(plots, "gid").run()

_plot(plots, neigh)
import matplotlib.pyplot as plt
ax = gdf.boundary.plot(color="grey")
plots.boundary.plot(ax=ax, color="red")
plt.show()
"""

"""
from geopandas import read_file
from t4gpd.demos.GeoDataFrameDemosA import GeoDataFrameDemosA
from t4gpd.morph.STVoronoiPartition2 import STVoronoiPartition2


bdtdir = "/home/tleduc/data/bdtopo/\
BDTOPO_3-3_TOUSTHEMES_SHP_LAMB93_D044_2024-03-15\
/BDTOPO/1_DONNEES_LIVRAISON_2024-04-00042/\
BDT_3-3_SHP_LAMB93_D044-ED2024-03-15"

roi = GeoDataFrameDemosA.hamainaP121roi()

gdf = read_file(f"{bdtdir}/BATI/BATIMENT.shp",
                mask=roi, engine="pyogrio")
gdf = gdf.clip(roi)
gdf["gid"] = range(len(gdf))

vor = STVoronoiPartition2(gdf, max_segment_length=0.25, erosion=0.5).run()
vor = vor.clip(roi)

neigh = STMakeNeighborhood(vor, "gid").run()
gdf.to_file("/tmp/abc.gpkg", layer="buildings")
vor.to_file("/tmp/abc.gpkg", layer="voronoi")
neigh.to_file("/tmp/abc.gpkg", layer="neighborhood")
"""



'''
Created on 27 sep. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import radians
from pandas import DataFrame, concat
from shapely import Point
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.proj.AEProjectionLib import AEProjectionLib
from t4gpd.resilientgaia.SatelliteLib import SatelliteLib


class STSatelliteOnSiteProjection(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, sensors, gid="gid", timestamp="timeUTC",
                 proj="Stereographic", size=1):
        '''
        Constructor
        '''
        if not isinstance(sensors, GeoDataFrame):
            raise IllegalArgumentTypeException(
                sensors, "GeoDataFrame")
        self.sensors = sensors
        self.gid, self.timestamp = gid, timestamp
        self.prj = AEProjectionLib.projection_switch(proj)
        self.size = size

    def __prj_aux(self, row):
        result = []
        for i in range(SatelliteLib.NSAT):
            az, el = row[f"sat_{i}_az"], row[f"sat_{i}_el"]
            if (az is None) or (el is None):
                result.append(None)
            else:
                az = AngleLib.northCW2eastCCW(az, degree=True)
                az, el = radians(az), radians(el)
                pp = Point(self.prj(row.geometry, az, el, self.size))
                sr = row[f"sat_{i}_sr"]
                pp = GeomLib.forceZCoordinateToZ0(pp, z0=sr)
                result.append(pp)
        return result

    def __prj(self):
        left = self.sensors.loc[:, [self.gid, self.timestamp, "geometry"]]
        left.reset_index(drop=True, inplace=True)

        right = self.sensors.apply(
            lambda row: self.__prj_aux(row),
            axis=1)
        right = right.to_frame(name="satellites")
        right = DataFrame(right.satellites.to_list(), columns=[
            f"sat_{i}" for i in range(SatelliteLib.NSAT)])

        df = concat([left, right], axis=1)
        df.dropna(axis=1, how="all", inplace=True)
        return df

    @ staticmethod
    def radar(sat, gid, constellations=["GPS", "Galileo", "GLONASS"], size=1, ofile=None):
        def __label_radar(geom):
            from shapely import get_num_points
            if 2 == get_num_points(geom):
                return 1
            return 52

        def __color_satellite(satName):
            match satName[0]:
                case "R":
                    return "red"
                case "E":
                    return "blue"
                case "G":
                    return "green"
            raise Exception("Unreachable instruction!")

        import matplotlib.pyplot as plt
        from matplotlib.colors import ListedColormap
        from numpy import linspace
        from shapely import LineString, get_point
        from shapely.affinity import rotate, translate
        from shapely.wkt import loads

        df = sat.loc[sat[sat.gid == gid].index]
        df.reset_index(inplace=True)

        vp = loads(df.loc[0, "viewpoint"])

        xaxis = LineString(
            [translate(vp, xoff=-size), translate(vp, xoff=size)])
        axes = GeoDataFrame([{
            "geometry": rotate(xaxis, angle=angle, origin=vp, use_radians=False),
            "label1": f"{270-angle}°",
            "label2": f"{AngleLib.northCW2eastCCW(angle, degree=True)}°"
        } for angle in range(0, 180, 30)])
        circles = GeoDataFrame([{
            "geometry": vp.buffer(size*buffdist/90).boundary,
            "label1": "",
            "label2": int(90 - buffdist)
        } for buffdist in linspace(15, 90, 6)])
        radar = concat([axes, circles])
        radar.reset_index(inplace=True)

        my_cmap = ListedColormap(["red", "green", "blue"])
        fig, ax = plt.subplots(figsize=(8.26, 8.26))
        radar.plot(ax=ax, color="grey", linestyle="dotted", linewidth=0.3)
        radar.apply(lambda x: ax.annotate(
            text=x.label1, xy=get_point(x.geometry, 0).coords[0],
            color="grey", size=10, ha="center", va="center"), axis=1)
        radar.apply(lambda x: ax.annotate(
            text=x.label2, xy=get_point(
                x.geometry, __label_radar(x.geometry)).coords[0],
            color="grey", size=10, ha="center"), axis=1)
        df.plot(ax=ax, marker="o", column="constellation", cmap=my_cmap)
        df.apply(lambda x: ax.annotate(
            text=x.satellite, xy=x.geometry.coords[0][:2],
            color=__color_satellite(x.satellite), size=12, ha="left"), axis=1)
        ax.axis("off")
        if ofile is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)

        return df, radar

    def run(self):
        df = self.__prj()
        df = df.melt(id_vars=[self.gid, self.timestamp, "geometry"])
        df["viewpoint"] = df.geometry.apply(lambda g: g.wkt)
        df.drop(columns=["geometry"], inplace=True)
        df.rename(columns={"variable": "satName",
                           "value": "geometry"}, inplace=True)
        df["satellite"] = df.satName.apply(
            lambda v: SatelliteLib.get_satellite_name(int(v[4:])))
        df["constellation"] = df.satellite.apply(
            lambda v: SatelliteLib.constellation(v))
        df = df.loc[df[df.geometry.apply(lambda g: g.is_valid)].index]
        df = df.set_crs(self.sensors.crs, allow_override=True)

        return df


"""
from datetime import timezone
from t4gpd.resilientgaia.STECEF2AERSatelliteReader import STECEF2AERSatelliteReader

ifile = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data/20240917_parcours_timestamp.csv"
aersat = STECEF2AERSatelliteReader(ifile, timestampFieldName="timeUTC").run()
aersat["gid"] = range(len(aersat))
aersat = aersat.set_index("timeUTC").tz_convert(timezone.utc).reset_index()
gids = [959]
aersat = aersat.loc[aersat[aersat.gid.isin(gids)].index]

# %run t4gpd/resilientgaia/STSatelliteOnSiteProjection.py
prjsat = STSatelliteOnSiteProjection(aersat, gid="gid",
                                     timestamp="timeUTC", proj="Stereographic", size=1).run()

STSatelliteOnSiteProjection.radar(prjsat, gids[0])
"""



'''
Created on 19 juin 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import MultiPoint, minimum_clearance, segmentize
from shapely.geometry import JOIN_STYLE
from shapely.ops import voronoi_diagram
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STVoronoiPartition2(object):
    '''
    classdocs
    '''

    def __init__(self, gdf, max_segment_length=1.0, erosion=0):
        '''
        Constructor
        '''
        def __erode(geom, erosion):
            if GeomLib.isPolygonal(geom):
                if ((0 == erosion) or (erosion > minimum_clearance(geom))):
                    _erosion = 0
                else:
                    _erosion = -erosion
                return geom.buffer(_erosion, join_style=JOIN_STYLE.mitre)
            return geom

        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        self.crs = gdf.crs
        self.gdf = gdf.copy(deep=True)
        self.gdf.geometry = self.gdf.geometry.apply(
            lambda g: __erode(g, erosion))

        _multiPoints = []
        for geom in self.gdf.geometry:
            _multiPoints += GeomLib.getListOfShapelyPoints(
                segmentize(geom, max_segment_length), withoutClosingLoops=False)
        _multiPoints = MultiPoint(_multiPoints)

        self.multiPoints = _multiPoints

    def run(self):
        vor = GeoDataFrame([{"geometry": v} for v in voronoi_diagram(
            self.multiPoints).geoms], crs=self.crs)
        vor = vor.sjoin(self.gdf, how="inner",
                        predicate="intersects", rsuffix="right01234567889")
        vor = vor.dissolve(by="index_right01234567889", as_index=False)
        vor.drop(columns="index_right01234567889", inplace=True)
        return vor


"""
def _plot(gdf, vor):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
    gdf.boundary.plot(ax=ax, color="grey", hatch="/", label="Input geoms")
    gdf.apply(lambda x: ax.annotate(
                text=x.gid, xy=x.geometry.centroid.coords[0],
                color="black", size=12, ha="center", va="center"), axis=1)
    vor.boundary.plot(ax=ax, color="red", label="Voronoi Tessel.")
    vor.apply(lambda x: ax.annotate(
                text=x.gid, xy=x.geometry.centroid.coords[0],
                color="red", size=12, ha="left", va="top"), axis=1)
    ax.legend(fontsize=14)
    fig.tight_layout()
    plt.show()
    plt.close(fig)
    
from shapely import box, MultiPolygon, Polygon
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos

gdf = GeoDataFrameDemos.regularGridOfPlots(nlines=3, ncols=3)
d = 1e-1
gdf = GeoDataFrame([
    # {"gid": 1, "geometry": MultiPolygon([box(0,0,1,2), box(2+d,0,4,2)])},
    {"gid": 1, "geometry": MultiPolygon([box(0,0,1,2)])},
    {"gid": 2, "geometry": Polygon([(1+d,1), (2,0), (2,2)])},
    {"gid": 3, "geometry": MultiPolygon([box(2+d,0,4,2)])},
])
vor = STVoronoiPartition2(gdf, max_segment_length=0.1).run()
_plot(gdf, vor)
"""

"""
from t4gpd.demos.GeoDataFrameDemosA import GeoDataFrameDemosA
from geopandas import read_file

def _plot(gdf, vor):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
    gdf.plot(ax=ax, color="grey", edgecolor="dimgrey",
             alpha=0.5, label="Input geoms")
    # gdf.apply(lambda x: ax.annotate(
    #     text=x.gid, xy=x.geometry.centroid.coords[0],
    #     color="black", size=12, ha="center", va="center"), axis=1)
    vor.boundary.plot(ax=ax, color="red", linewidth=0.5,
                      label="Voronoi Tessel.")
    # vor.apply(lambda x: ax.annotate(
    #     text=x.gid, xy=x.geometry.centroid.coords[0],
    #     color="red", size=12, ha="left", va="top"), axis=1)
    ax.axis("off")
    ax.legend(fontsize=14)
    fig.tight_layout()
    plt.show()
    plt.close(fig)


bdtdir = "/home/tleduc/data/bdtopo/\
BDTOPO_3-3_TOUSTHEMES_SHP_LAMB93_D044_2024-03-15\
/BDTOPO/1_DONNEES_LIVRAISON_2024-04-00042/\
BDT_3-3_SHP_LAMB93_D044-ED2024-03-15"

roi = GeoDataFrameDemosA.hamainaP121roi()

gdf = read_file(f"{bdtdir}/BATI/BATIMENT.shp",
                mask=roi, engine="pyogrio")
gdf = gdf.clip(roi)
gdf["gid"] = range(len(gdf))

vor = STVoronoiPartition2(gdf, max_segment_length=0.25, erosion=0.5).run()
vor = vor.clip(roi)

_plot(gdf, vor)
"""



'''
Created on 22 aug. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import MultiLineString, Point, get_coordinates, get_parts
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.SphericalCartesianCoordinates import SphericalCartesianCoordinates
from t4gpd.commons.raycasting.PrepareMasksLib import PrepareMasksLib


class SurroundingMasks25DLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def __c2s(viewpoint, mask):
        a, b = get_coordinates(mask, include_z=True)
        ar, alon, alat = SphericalCartesianCoordinates.fromCartesianToSphericalCoordinates(
            a[0] - viewpoint.x, a[1] - viewpoint.y, a[2] - viewpoint.z)
        br, blon, blat = SphericalCartesianCoordinates.fromCartesianToSphericalCoordinates(
            b[0] - viewpoint.x, b[1] - viewpoint.y, b[2] - viewpoint.z)
        assert (alat >= 0), "'alat' must be positive or zero"
        assert (blat >= 0), "'blat' must be positive or zero"
        alon, alat = AngleLib.toDegrees(
            alon), AngleLib.toDegrees(alat)  # DEBUG
        blon, blat = AngleLib.toDegrees(
            blon), AngleLib.toDegrees(blat)  # DEBUG
        # print(f"{alon=:.1f} --- {blon=:.1f}")  # DEBUG
        return ar, alon, alat, br, blon, blat

    @staticmethod
    def __isHiddenBy(viewpoint, curr, other):
        # Returns True if "other" is hidden by "curr" when observing from "viewpoint"
        cr1, clon1, clat1 = curr["__r1__"], curr["__lon1__"], curr["__lat1__"]
        cr2, clon2, clat2 = curr["__r2__"], curr["__lon2__"], curr["__lat2__"]
        or1, olon1, olat1 = other["__r1__"], other["__lon1__"], other["__lat1__"]
        or2, olon2, olat2 = other["__r2__"], other["__lon2__"], other["__lat2__"]

        return ((clon1 >= olon1 >= olon2 >= clon2) and
                (min(clat1, clat2) >= max(olat1, olat2)) and
                (max(cr1, cr2) < min(or1, or2)))

    @ staticmethod
    def removeHidden25DMasks(vp_and_masks):
        if not isinstance(vp_and_masks, GeoDataFrame):
            raise IllegalArgumentTypeException(vp_and_masks, "GeoDataFrame")
        vp_and_masks2 = vp_and_masks.explode(ignore_index=True)

        vp_and_masks2 = vp_and_masks2.to_dict(orient="index")
        for key in vp_and_masks2.keys():
            viewpoint = vp_and_masks2[key]["__VIEWPOINT_GEOM__"]
            mask = vp_and_masks2[key]["geometry"]
            tmp = SurroundingMasks25DLib.__c2s(viewpoint, mask)
            vp_and_masks2[key]["__r1__"] = tmp[0]
            vp_and_masks2[key]["__lon1__"] = tmp[1]
            vp_and_masks2[key]["__lat1__"] = tmp[2]
            vp_and_masks2[key]["__r2__"] = tmp[3]
            vp_and_masks2[key]["__lon2__"] = tmp[4]
            vp_and_masks2[key]["__lat2__"] = tmp[5]

        before = len(vp_and_masks2)  # DEBUG

        # Remove completely hidden edges (phase 1)
        hiddenEdgeIds = set()
        for i, curr in vp_and_masks2.items():
            if (not i in hiddenEdgeIds):
                for j, other in vp_and_masks2.items():
                    if (not j in hiddenEdgeIds) and (i != j):
                        if SurroundingMasks25DLib.__isHiddenBy(viewpoint, curr, other):
                            hiddenEdgeIds.add(j)

        for i in sorted(hiddenEdgeIds, reverse=True):
            del (vp_and_masks2[i])
        print(f"{before} --> {len(vp_and_masks2)}")  # DEBUG

        vp_and_masks2 = GeoDataFrame(
            vp_and_masks2.values(), crs=vp_and_masks.crs)
        vp_and_masks2.drop(columns=["__r1__", "__lon1__", "__lat1__",
                                    "__r2__", "__lon2__", "__lat2__"],
                           inplace=True)
        return vp_and_masks2

    @ staticmethod
    def test(ofile=None):
        import matplotlib.pyplot as plt
        from matplotlib_scalebar.scalebar import ScaleBar
        from shapely import box
        from shapely.wkt import loads
        from t4gpd.demos.GeoDataFrameDemosC import GeoDataFrameDemosC

        buildings = GeoDataFrameDemosC.irisTastavinBuildings()
        viewpoints = GeoDataFrame([{
            "gid": 123,
            "geometry": loads("Point (770402.89999999979045242 6277956.6500000013038516 0)")
        }], crs=buildings.crs)

        vp_and_masks1 = PrepareMasksLib.removeNonVisible25DMasks(
            viewpoints, buildings, "HAUTEUR", horizon=100.0, h0=0.0,
            encode=True
        )
        vp_and_masks1.__VIEWPOINT_GEOM__ = vp_and_masks1.__VIEWPOINT_GEOM__.apply(
            lambda v: loads(v))
        vp_and_masks2 = SurroundingMasks25DLib.removeHidden25DMasks(
            vp_and_masks1)

        # MAPPING
        minx, miny, maxx, maxy = box(
            *viewpoints.total_bounds).buffer(30.0).bounds
        minx, miny, maxx, maxy = vp_and_masks1.total_bounds

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.35 * 8.26))
        viewpoints.plot(ax=ax, color="green", marker="P")
        buildings.plot(ax=ax, color="lightgrey")
        vp_and_masks1.plot(ax=ax, color="red", linewidth=5)
        vp_and_masks2.plot(ax=ax, color="blue", linewidth=2)
        ax.axis("off")
        ax.axis([minx, maxx, miny, maxy])
        scalebar = ScaleBar(dx=1.0, units="m", length_fraction=None, box_alpha=0.85,
                            width_fraction=0.005, location="lower right", frameon=True)
        ax.add_artist(scalebar)
        fig.tight_layout()
        plt.show()
        plt.close(fig)

        if not ofile is None:
            buildings.to_file(ofile, layer="buildings")
            viewpoints.to_file(ofile, layer="viewpoints")
            vp_and_masks1.to_file(ofile, layer="vp_and_masks1")
            vp_and_masks2.to_file(ofile, layer="vp_and_masks2")

        return vp_and_masks1, vp_and_masks2


# vp_and_masks1, vp_and_masks2 = SurroundingMasks25DLib.test("/tmp/SurroundingMasks25DLib.gpkg")
